import rebound
from ctypes import c_double, byref
import unittest
from time import sleep

class TestRandom(unittest.TestCase):
    def test_uniform(self):
        rebound.clibrebound.reb_random_uniform.restype = c_double
        simulation = rebound.Simulation()
        for simp in [0, byref(simulation)]:
            for i in range(10):
                r = rebound.clibrebound.reb_random_uniform(simp, c_double(0.0), c_double(1.0))
                self.assertLess(r, 1.0)
                self.assertGreater(r, 0.0)
    
    def test_powerlaw(self):
        rebound.clibrebound.reb_random_powerlaw.restype = c_double
        simulation = rebound.Simulation()
        for simp in [0, byref(simulation)]:
            for slope in [-1.0, 0.0, 1.0]:
                for i in range(10):
                    r = rebound.clibrebound.reb_random_powerlaw(simp, c_double(1.0), c_double(2.0), c_double(slope))
                    self.assertLess(r, 2.0)
                    self.assertGreater(r, 1.0)
    
    def test_normal(self):
        rebound.clibrebound.reb_random_normal.restype = c_double
        simulation = rebound.Simulation()
        for simp in [0, byref(simulation)]:
            for i in range(10):
                r = rebound.clibrebound.reb_random_normal(simp, c_double(1.0))
                self.assertLess(r, 1e5)
                self.assertGreater(r, -1e5)
    
    def test_rayleigh(self):
        rebound.clibrebound.reb_random_rayleigh.restype = c_double
        simulation = rebound.Simulation()
        for simp in [0, byref(simulation)]:
            for i in range(10):
                r = rebound.clibrebound.reb_random_rayleigh(simp, c_double(1.0))
                self.assertLess(r, 1e5)
                self.assertGreater(r, 0.0)

    def test_reproducible(self):
        rebound.clibrebound.reb_random_uniform.restype = c_double
        sim1 = rebound.Simulation()
        sim2 = sim1.copy()
        sleep(0.05) # Windows implementation is very slow, so we wait to get a different seed
        sim3 = rebound.Simulation()
        self.assertEqual(sim1.rand_seed, sim2.rand_seed)
        self.assertNotEqual(sim2.rand_seed, sim3.rand_seed)
        r1 = rebound.clibrebound.reb_random_uniform(byref(sim1), c_double(0.0), c_double(1.0))
        r2 = rebound.clibrebound.reb_random_uniform(byref(sim2), c_double(0.0), c_double(1.0))
        r3 = rebound.clibrebound.reb_random_uniform(byref(sim3), c_double(0.0), c_double(1.0))
        self.assertEqual(r1, r2)
        self.assertNotEqual(r2, r3)
        r2 = rebound.clibrebound.reb_random_uniform(byref(sim2), c_double(0.0), c_double(1.0))
        r3 = rebound.clibrebound.reb_random_uniform(byref(sim3), c_double(0.0), c_double(1.0))
        self.assertNotEqual(r2, r3)
    
if __name__ == "__main__":
    unittest.main()



from ctypes import c_uint, c_double, c_void_p, CFUNCTYPE, POINTER, c_int, Structure
from .. import clibrebound

class ODE(Structure):
    @property
    def derivatives(self):
        raise AttributeError("You can only set C function pointers from python.")
    @derivatives.setter
    def derivatives(self, func):
        self._dfp = ODEDER(func)
        func.argtypes = self._dfp.argtypes # I do not understand why this is needed
        self._derivatives = self._dfp
    def update_particles(self):
        clibrebound.reb_integrator_bs_update_particles(self.r, None) 

from ..simulation import Simulation
ODE._fields_ = [
                ("length", c_uint),
                ("y", POINTER(c_double)),
                ("needs_nbody", c_uint),
                ("ref", c_void_p),
                ("_derivatives", CFUNCTYPE(None,POINTER(ODE), POINTER(c_double), POINTER(c_double), c_double)),
                ("_getscale", CFUNCTYPE(None,POINTER(ODE), POINTER(c_double), POINTER(c_double))),
                ("_pre_timestep", CFUNCTYPE(None,POINTER(ODE), POINTER(c_double))),
                ("_post_timestep", CFUNCTYPE(None,POINTER(ODE), POINTER(c_double))),
                ("N_allocated", c_uint),
                ("_scale", POINTER(c_double)),
                ("_C", POINTER(c_double)),
                ("_D", POINTER(POINTER(c_double))),
                ("_y1", POINTER(c_double)),
                ("_y0Dot", POINTER(c_double)),
                ("_yDot", POINTER(c_double)),
                ("_yTmp", POINTER(c_double)),
                ("r", POINTER(Simulation)),
            ]               

ODEDER = CFUNCTYPE(None,POINTER(ODE), POINTER(c_double), POINTER(c_double), c_double)

class IntegratorBS(Structure):
    """
    This class is an abstraction of the C-struct reb_integrator_bs.
    It controls the behaviour of the Gragg-Bulirsch-Stoer integrator.
    """
    _fields_ = [
                ("eps_abs", c_double),
                ("eps_rel", c_double),
                ("min_dt", c_double),
                ("max_dt", c_double),
                ("_nbody_ode", POINTER(ODE)),
                ("_sequence", POINTER(c_int)),
                ("_cost_per_step", POINTER(c_int)),
                ("_cost_per_time_unit", POINTER(c_double)),
                ("_optimal_step", POINTER(c_double)),
                ("_coeff", POINTER(c_double)),
                ("dt_proposed", c_double),
                ("_first_or_last_step", c_int),
                ("_previous_rejected", c_int),
                ("_target_iter", c_int),
                ("_user_ode_needs_nbody", c_int),
            ]               



import rebound
import rebound.binary_field_descriptor
import unittest

class TestBinaryFields(unittest.TestCase):
    def test_unique(self):
        l = rebound.binary_field_descriptor.binary_field_descriptor_list()
        types = []
        names = []
        for i in l:
            types.append(i.type)
            names.append(i.name)
        self.assertEqual(len(types), len(set(types)))
        self.assertEqual(len(names), len(set(names)))


if __name__ == "__main__":
    unittest.main()




import ctypes

from .. import clibrebound
from ..simulation import Simulation
from ..particle import Particle
from ..vectors import Vec3dBasic

class IntegratorMercurius(ctypes.Structure):
    """
    This class is an abstraction of the C-struct reb_integrator_mercurius.
    It controls the behaviour of the MERCURIUS integrator.  See Rein et al. (2019) 
    for more details.
    
    :ivar float r_crit_hill:      
        Switching radius in units of the hill radius.

    Example usage:
    
    >>> sim = rebound.Simulation()
    >>> sim.integrator = "mercurius"
    >>> sim.ri_mercurius.r_crit_hill = 3.

    """
    def __repr__(self):
        return '<{0}.{1} object at {2}, safe_mode={3}, is_synchronized={4}, r_crit_hill={5}>'.format(self.__module__, type(self).__name__, hex(id(self)), self.safe_mode, self.is_synchronized, self.r_crit_hill)

    _fields_ = [("_L", ctypes.CFUNCTYPE(ctypes.c_double, ctypes.POINTER(Simulation), ctypes.c_double, ctypes.c_double)),
                ("r_crit_hill", ctypes.c_double),
                ("recalculate_coordinates_this_timestep", ctypes.c_uint),
                ("recalculate_r_crit_this_timestep", ctypes.c_uint),
                ("safe_mode", ctypes.c_uint),
                ("is_synchronized", ctypes.c_uint),
                ("mode", ctypes.c_uint),
                ("_encounter_N", ctypes.c_uint),
                ("_encounter_N_active", ctypes.c_uint),
                ("_tponly_encounter", ctypes.c_uint),
                ("_N_allocated", ctypes.c_uint),
                ("_N_allocated_additional_forces", ctypes.c_uint),
                ("_N_allocated_dcrit", ctypes.c_uint),
                ("_dcrit", ctypes.POINTER(ctypes.c_double)),
                ("_particles_backup", ctypes.POINTER(Particle)),
                ("_particles_backup_additional_forces", ctypes.POINTER(Particle)),
                ("_encounter_map", ctypes.POINTER(ctypes.c_int)),
                ("_com_pos", Vec3dBasic),
                ("_com_vel", Vec3dBasic),
                ]
    @property
    def L(self):
        raise AttributeError("You can only set C function pointers from python.")
    @L.setter
    def L(self, func):
        if func == "mercury":
            self._L = ctypes.cast(clibrebound.reb_integrator_mercurius_L_mercury,MERCURIUSLF)
        elif func == "C4":
            self._L = ctypes.cast(clibrebound.reb_integrator_mercurius_L_C4,MERCURIUSLF)
        elif func == "C5":
            self._L = ctypes.cast(clibrebound.reb_integrator_mercurius_L_C5,MERCURIUSLF)
        elif func == "infinity":
            self._L = ctypes.cast(clibrebound.reb_integrator_mercurius_L_infinity,MERCURIUSLF)
        else:
            self._Lfp = MERCURIUSLF(func)
            self._L = self._Lfp

MERCURIUSLF = ctypes.CFUNCTYPE(ctypes.c_double, ctypes.POINTER(Simulation), ctypes.c_double, ctypes.c_double)



import rebound
import unittest

class TestSerialize(unittest.TestCase):
    def setUp(self):
        self.sim = rebound.Simulation()
        self.sim.t = 1.246
        self.sim.add(m=1.)
        self.sim.add(a=1.)
    
    def tearDown(self):
        self.sim = None

    def test_serialize(self):
        try:
            import numpy as np
        except:
            # Make numpy tests optional
            print("WARNING: Not testing serialization because numpy is not available")
            return;

        a = np.zeros((self.sim.N,3),dtype="float64")
        self.sim.serialize_particle_data(xyz=a)

        self.assertEqual(a[1][0],1)
        self.assertEqual(a[1][1],0)
        self.assertEqual(a[1][2],0)

        self.sim.particles[0].xyz = [0,0,0]

        b = np.zeros(self.sim.N,dtype="uint32")
        c = np.zeros(self.sim.N)
        self.sim.serialize_particle_data(r=c,hash=b)

        with self.assertRaises(AttributeError):
            self.sim.serialize_particle_data(r=b)

        with self.assertRaises(AttributeError):
            self.sim.serialize_particle_data(hash=a)
        
        with self.assertRaises(AttributeError):
            self.sim.serialize_particle_data(xyz=c)

    
    def test_set_serialize(self):
        try:
            import numpy as np
        except:
            # Make numpy tests optional
            print("WARNING: Not testing serialization because numpy is not available")
            return;
        
        for i in range(self.sim.N):
            self.sim.particles[i].xyz = [0,0,0]
            self.assertEqual(self.sim.particles[i].x,0)
            self.assertEqual(self.sim.particles[i].y,0)
            self.assertEqual(self.sim.particles[i].z,0)
        
        a = np.zeros((self.sim.N,3),dtype="float64")
        a[0][1] = 2
        a[1][2] = 6
        self.sim.set_serialized_particle_data(xyz=a)
        self.assertEqual(self.sim.particles[0].y,2)
        self.assertEqual(self.sim.particles[1].z,6)

    
if __name__ == "__main__":
    unittest.main()



from ctypes import c_double, byref
import rebound

def mod2pi(x):
    try:
        x = float(x)
    except:
        ValueError("Argument of mod2pi needs to be a float.")
    clibrebound.reb_mod2pi.restype = c_double
    return clibrebound.reb_mod2pi(c_double(x))

def M_to_f(e, M):
    try:
        e = float(e)
        M = float(M)
    except:
        ValueError("Arguments of M_to_f need to be floats.")
    clibrebound.reb_M_to_f.restype = c_double
    return clibrebound.reb_M_to_f(c_double(e), c_double(M))

def E_to_f(e, E):
    try:
        e = float(e)
        E = float(E)
    except:
        ValueError("Arguments of E_to_f need to be floats.")
    clibrebound.reb_E_to_f.restype = c_double
    return clibrebound.reb_E_to_f(c_double(e), c_double(E))

def M_to_E(e, M):
    try:
        e = float(e)
        M = float(M)
    except:
        ValueError("Arguments of M_to_E need to be floats.")
    clibrebound.reb_M_to_E.restype = c_double
    return clibrebound.reb_M_to_E(c_double(e), c_double(M))

def spherical_to_xyz(magnitude=1., theta=0., phi=0.):
    """Initialize Cartesian vector from its magnitude and two spherical angles theta (polar angle measured from z) and phi (azimuthal angle measured from x)

    Arguments
    ---------
    magnitude: float
        Magnitude of the vector
    theta: float
        Polar angle of the vector, i.e. measured from the z axis
    phi: float
        Azimuthal angle of the vector, i.e. measured counterclockwise from the x axis

    Returns
    -------
    List of [x,y,z] components
    """ 
    clibrebound.reb_tools_spherical_to_xyz.restype = rebound.Vec3dBasic
    xyz = clibrebound.reb_tools_spherical_to_xyz(c_double(magnitude), c_double(theta), c_double(phi))
    return rebound.Vec3d(xyz)

def xyz_to_spherical(vector):
    """Return spherical angle theta measured from z axis
    
    Arguments
    ---------
    vector: List-like (e.g., list, numpy array)
        3D Cartesian vector to convert to spherical coordinates

    Returns
    -------
    List [magnitude, theta, phi] of vector magnitude, polar angle theta (from z axis) and azimuthal angle phi (from x axis)
    """
    magnitude = c_double()
    theta = c_double()
    phi = c_double()
    clibrebound.reb_tools_xyz_to_spherical(rebound.Vec3d(vector)._vec3d, byref(magnitude), byref(theta), byref(phi))
    return magnitude.value, theta.value, phi.value

from . import clibrebound



import sys
from ctypes import c_uint32, c_uint, c_uint64, addressof, POINTER, pointer, byref
from .hash import hash as rebhash
from .particle import Particle
from . import clibrebound, ParticleNotFound

try:
    # Required for Python>=3.9
    from collections.abc import MutableMapping
except:
    from collections import MutableMapping


class Particles(MutableMapping):
    """
    This class allows the user to access particles like a dictionary using the particle's 1) index 2) hash 3) string (which will be converted to hash).
    Allows for negative indices and slicing.
    """
    def __init__(self, sim):
        self.sim = sim

    @property
    def _ps(self):
        ParticleList = Particle*self.sim.N
        pl = ParticleList.from_address(addressof(self.sim._particles.contents))
        pl._sim = self.sim # keep reference to sim until ParticleList is deallocated to avoid memory issues
        return pl

    def __getitem__(self, key):
        hash_types = c_uint32, c_uint, c_uint64
        PY3 = sys.version_info[0] == 3
        if PY3:
            string_types = str,
            int_types = int,
        else:
            string_types = basestring,
            int_types = int, long,
       
        if isinstance(key, slice):
            return [self[i] for i in range(*key.indices(len(self)))]

        if isinstance(key, int_types):
            if key < 0: # accept negative indices
                key += self.sim.N
            if key < 0 or key >= self.sim.N:
                raise AttributeError("Index {0} used to access particles out of range.".format(key))
            return self._ps[key]

        else:
            clibrebound.reb_simulation_particle_by_hash.restype = POINTER(Particle)
            if isinstance(key, string_types):
                key = rebhash(key)
            elif not isinstance(key, hash_types):
                raise AttributeError("Expecting string, integer or ctypes.c_uint32 as argument to sim.particles.  See UniquelyIdentifyingParticlesWithHashes.ipynb ipython_example.")

            ptr = clibrebound.reb_simulation_particle_by_hash(byref(self.sim), key) 
            
            if ptr:
                p = Particle
                return p.from_address(addressof(ptr.contents))
            else:
                raise ParticleNotFound("Particle was not found in the simulation.") 

    def __setitem__(self, key, value):
        if isinstance(value, Particle):
            value._sim = pointer(self.sim)
            p = self[key]
            if p.index == -1:
                raise AttributeError("Can't set particle (particle not found in simulation).")
            else:
                self._ps[p.index] = value

    def __delitem__(self, key):
        pass

    def __iter__(self):
        if self.sim.N>0:
            for p in self._ps:
                yield p

    def __len__(self):
        return self.sim.N



import rebound
import unittest
import pickle
import warnings

class TestPickle(unittest.TestCase):
    def test_pickle_basic(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1,a=1)
        sim.step()
        
        with open("test.pickle", "wb") as f:
            pickle.dump(sim, f)

        with open("test.pickle", "rb") as f:
            sim2 = pickle.load(f)

        self.assertEqual(sim,sim2)
        sim.step()
        sim2.step()
        self.assertEqual(sim,sim2)
    
    def test_pickle_warning(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.collision_resolve = "merge"
        sim.step()
        
        with open("test.pickle", "wb") as f:
            pickle.dump(sim, f)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            with open("test.pickle", "rb") as f:
                sim2 = pickle.load(f)

            self.assertEqual(1, len(w)) 
        self.assertNotEqual(sim,sim2)
        sim2.collision_resolve = "merge"
        sim.step()
        sim2.step()
        self.assertEqual(sim,sim2)

    def test_pickle_particle(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1,a=1)
        sim.step()
        
        with open("test.pickle", "wb") as f:
            pickle.dump(sim.particles[1], f)

        with open("test.pickle", "rb") as f:
            p2 = pickle.load(f)

        self.assertEqual(sim.particles[1], p2)

        # Pointers are set to zero when unpickling
        self.assertNotEqual(sim.particles[1]._sim, p2._sim)
        self.assertEqual(p2.sim, 0)

if __name__ == "__main__":
    unittest.main()



import ctypes

class ParticleInt(ctypes.Structure):
    """ Used for Janus integrator only """
    _fields_ = [
                ("x", ctypes.c_int64),
                ("y", ctypes.c_int64),
                ("z", ctypes.c_int64),
                ("vx", ctypes.c_int64),
                ("vy", ctypes.c_int64),
                ("vz", ctypes.c_int64),
                ]

class IntegratorJanus(ctypes.Structure):
    _fields_ = [
                ("scale_pos",ctypes.c_double),
                ("scale_vel",ctypes.c_double),
                ("order", ctypes.c_uint),
                ("recalculate_integer_coordinates_this_timestep", ctypes.c_uint),
                ("p_int", ctypes.POINTER(ParticleInt)),
                ("_N_allocated",ctypes.c_uint),
                ]



from ctypes import Structure, c_double

class Vec6d(Structure):
    _fields_ = [("x", c_double),
                ("y", c_double),
                ("z", c_double),
                ("vx", c_double),
                ("vy", c_double),
                ("vz", c_double)]

class Vec3dBasic(Structure):
    """
    Internal use only. Not used as Vec3d directly because assigments to numpy arrays don't worl
    """
    _fields_ = [("x", c_double),
                ("y", c_double),
                ("z", c_double)]

class Vec3d:
    """
    Class for 3D Cartesian vectors. 
    """
    _vec3d = None

    @property
    def __array_interface__(self):
        return {"shape": (3,), "typestr": "<f8", "data": self._vec3d} 

    def __init__(self, *args):
        if len(args) == 1: 
            vec = args[0]
            if isinstance(vec,Vec3dBasic):
                vec = [vec.x, vec.y, vec.z]
            elif isinstance(vec,str):
                vec = vec.lower()
                if vec != "x" and vec !="y" and vec != "z":
                    raise ValueError("When passing a string to create a Vec3D, it needs to be one of 'x', 'y', or  'z'")
                if vec == "x":
                    vec = [1,0,0]
                elif vec == "y":
                    vec = [0,1,0]
                if vec == "z":
                    vec = [0,0,1]
            else:
                vec = [float(vec[0]), float(vec[1]), float(vec[2])]
        elif len(args) >= 3:
            vec = [float(args[0]), float(args[1]), float(args[2])]
        self._vec3d =Vec3dBasic(vec[0],vec[1],vec[2])

    def __mul__(self, other):
        try:
            return Vec3d([self.x*other, self.y*other, self.z*other]) 
        except:
            return NotImplemented

    def __truediv__(self, other):
        if other==0.:
            raise ZeroDivisionError
        try:
            return Vec3d([self.x/other, self.y/other, self.z/other])
        except:
            return NotImplemented

    def __add__(self, other):
        try:
            o = Vec3d(other)
            return Vec3d([self[0]+other[0], self[1]+other[1], self[2]+other[2]]) 
        except:
            return NotImplemented
    
    def __sub__(self, other):
        try:
            o = Vec3d(other)
            return Vec3d([self[0]-other[0], self[1]-other[1], self[2]-other[2]]) 
        except:
            return NotImplemented
    
    def rotate(self, q):
        if not isinstance(q, Rotation):
            raise NotImplementedError
        clibrebound.reb_vec3d_irotate(byref(_vec3d), q)
        return self

    def normalize(self):
        clibrebound.reb_vec3d_normalize.restype = Vec3dBasic
        r = clibrebound.reb_vec3d_normalize(self._vec3d)
        self._vec3d = r._vec3d
        return self

    def __getitem__(self, key):
        if not isinstance(key, int):
            raise IndexError("Index must be an integer.")
        if key < 0 or key >= 3:
            raise IndexError("Vec3d has exactly three elements and can therefore not access the item with index "+str(key)+".")
        if key == 0:
            return self._vec3d.x
        if key == 1:
            return self._vec3d.y
        if key == 2:
            return self._vec3d.z

    def __setitem__(self, key, value):
        if not isinstance(key, int):
            raise IndexError("Index must be an integer.")
        if key < 0 or key >= 3:
            raise IndexError("Vec3d has exactly three elements and can therefore not access the item with index "+str(key)+".")
        if key == 0:
            self._vec3d.x = c_double(value)
        if key == 1:
            self._vec3d.y = c_double(value)
        if key == 2:
            self._vec3d.z = c_double(value)

    @property
    def x(self):
         return self._vec3d.x
    @x.setter
    def x(self, v):
         self._vec3d.x = v
    @property
    def y(self):
         return self._vec3d.y
    @y.setter
    def y(self, v):
         self._vec3d.y = v
    @property
    def z(self):
         return self._vec3d.z
    @z.setter
    def z(self, v):
         self._vec3d.z = v

    def __repr__(self):
        return '<{0}.{1} object at {2}, [{3}, {4}, {5}]>'.format(self.__module__, type(self).__name__, hex(id(self)), self._vec3d.x, self._vec3d.y, self._vec3d.z)
    



from ctypes import Structure, c_double, POINTER, c_uint32, c_int, c_uint, c_int64, c_uint64, c_void_p, c_char_p, CFUNCTYPE, byref, create_string_buffer, addressof, c_char, c_size_t, string_at, sizeof 
from . import clibrebound, Escape, NoParticles, Encounter, Collision, GenericError 
from .citations import cite
from .units import units_convert_particle, check_units, convert_G, hash_to_unit
from .hash import hash as rebhash, HashPointerPair
from .vectors import Vec3d, Vec3dBasic, Vec6d
import os
import sys
import warnings

# Avoid unboxing strings. See
# https://sourceforge.net/p/ctypes/mailman/message/8469497/
class allocated_c_char_p(c_char_p):
    pass

import types
      
### The following enum and class definitions need to
### consitent with those in rebound.h
        
INTEGRATORS = {"ias15": 0, "whfast": 1, "sei": 2, "leapfrog": 4, "none": 7, "janus": 8, "mercurius": 9, "saba": 10, "eos": 11, "bs": 12, "whfast512":21, "trace":25}
BOUNDARIES = {"none": 0, "open": 1, "periodic": 2, "shear": 3}
GRAVITIES = {"none": 0, "basic": 1, "compensated": 2, "tree": 3, "mercurius": 4, "jacobi": 5, "trace": 6}
COLLISIONS = {"none": 0, "direct": 1, "tree": 2, "line": 4, "linetree": 5}
# Format: Majorerror, id, message
BINARY_WARNINGS = [
    (True,  1, "Cannot read binary file. Check filename and file contents."),
    (False, 2, "Binary file was saved with a different version of REBOUND. Binary format might have changed."),
    (False, 4, "You have to reset function pointers after creating a reb_simulation struct with a binary file."),
    (False, 8, "Binary file might be corrupted. Number of particles found does not match particle number expected."),
    (True,  16, "Error while reading binary file (file was closed).",),
    (True,  32, "Index out of range.",),
    (True,  64, "Error while trying to seek file.",),
    (False, 128, "Encountered unknown field in file. File might have been saved with a different version of REBOUND."),
    (True,  256, "Integrator type is not supported by this simulationarchive version."),
    (False,  512, "The binary file seems to be corrupted. An attempt has been made to read the uncorrupted parts of it."),
    (True, 1024, "Reading old Simulationarchives (version < 2) is no longer supported. If you need to read such an archive, use a REBOUND version <= 3.26.3"),
]

# Note: name conflict with exception "Collision"
class CollisionS(Structure):
    _fields_ = [("p1", c_int),
                ("p2", c_int),
                ("gb", Vec6d),
                ("ri", c_int)]
    
    def __repr__(self):
        return '<{0}.{1} object at {2}, p1={3}, p2={4}>'.format(self.__module__, type(self).__name__, hex(id(self)), self.p1, self.p2)
    


class Simulation(Structure):
    """
    This is the REBOUND Simulation Class.
    In encapsulated an entire REBOUND simulation and is an abstraction of the C struct reb_simulation.

    ### Examples
    
    Most simulation parameters can be directly changed with the property syntax:

    >>> sim = rebound.Simulation()
    >>> sim.G = 1.                  # Sets the graviational constant (default 1)
    >>> sim.softening = 1.          # Sets the graviational softening parameter (default 0)
    >>> sim.testparticle_type = 1   # Allows massive particles to feel influence from testparticles (default 0)
    >>> sim.dt = 0.1                # Sets the timestep (will change for adaptive integrators such as IAS15).
    >>> sim.t = 0.                  # Sets the current simulation time (default 0)
    >>> print(sim.N)                # Gets the current number of particles
    >>> print(sim.N_active)         # Gets the current number of active particles
       
    By calling rebound.Simulation() as shown above, you create a new simulation object
    The following example creates a simulation, saves it to a file and then creates
    a copy of the simulation stored in the binary file.

    >>> sim = rebound.Simulation()
    >>> sim.add(m=1.)
    >>> sim.add(m=1.e-3,x=1.,vy=1.)
    >>> sim.save_to_file("simulation.bin")
    >>> sim_copy = rebound.Simulation("simulation.bin")
    
    Similarly, you can create a simulation from a Simulationarchive
    by specifying the snapshot you want to load. 

    >>> sim = rebound.Simulation("archive.bin", 34)
    
    or 
    
    >>> sim = rebound.Simulation(filename="archive.bin", snapshot=34)

    Finally, you can also create a new Simulation by passing a bytes object of a 
    Simulationarchive to Simulation():
    
    >>> sim = rebound.Simulation(open("archive.bin","rb").read())

    """
    def __new__(cls, *args, **kw):
        # Create a new simulation if no arguments given
        if len(args)==0:
            sim = super(Simulation,cls).__new__(cls)
            clibrebound.reb_simulation_init(byref(sim))
            return sim
        
        # If first argument is of type bytes, then unpickle a Simulation
        if isinstance(args[0], bytes):
            l = len(args[0]) 
            buft = c_char * l
            buf = buft.from_buffer_copy(args[0])
            # Note: Not calling Simulationarchive.
            # Doing this manually here because we need to keep the reference to buf.
            # So we can later access the contents of the archive to get the simulation.
            sa = Simulationarchive.__new__(Simulationarchive, None, None)
            w = c_int(0)
            clibrebound.reb_simulationarchive_init_from_buffer_with_messages(byref(sa), byref(buf), c_size_t(l), None, byref(w))
            sim = super(Simulation,cls).__new__(cls)
            clibrebound.reb_simulation_init(byref(sim))
            clibrebound.reb_simulation_create_from_simulationarchive_with_messages(byref(sim),byref(sa),c_int64(-1),byref(w))
            for majorerror, value, message in BINARY_WARNINGS:
                if w.value & value:
                    if majorerror:
                        raise RuntimeError(message)
                    else:  
                        # Just a warning
                        warnings.warn(message, RuntimeWarning)
            return sim
       
        # Create simulation from Simulationarchive
        if isinstance(args[0], Simulationarchive):
            sa = args[0]
        else:
            # Otherwise assume first argument is filename
            filename = args[0]
            if "filename" in kw:
                filename = kw["filename"]
            sa = Simulationarchive(filename,process_warnings=False)

        snapshot = -1
        if len(args)>1:
            snapshot = args[1]
        if "snapshot" in kw:
            snapshot = kw["snapshot"]

        if sa is not None:
            # Recreate exisitng simulation 
            sim = super(Simulation,cls).__new__(cls)
            clibrebound.reb_simulation_init(byref(sim))
            w = sa.warnings # warnings will be appended to previous warnings (as to not repeat them) 
            clibrebound.reb_simulation_create_from_simulationarchive_with_messages(byref(sim),byref(sa),c_int64(snapshot),byref(w))
            for majorerror, value, message in BINARY_WARNINGS:
                if w.value & value:
                    if majorerror:
                        raise RuntimeError(message)
                    else:  
                        # Just a warning
                        warnings.warn(message, RuntimeWarning)
            return sim

        # Still here? Then an error occured.
        raise RuntimeError("Can not create Simulation.")

    def __init__(self,filename=None,snapshot=None):
        self.save_messages = 1 # Warnings will be checked within python

    def __repr__(self):
        return '<{0}.{1} object at {2}, N={3}, t={4}>'.format(self.__module__, type(self).__name__, hex(id(self)), self.N, self.t)
    
    @classmethod
    def from_simulationarchive(cls, simulationarchive, snapshot=-1):
        return cls(filename=filename,snapshot=snapshot)

    @classmethod
    def from_file(cls, filename, snapshot=-1):
        return cls(filename=filename, snapshot=snapshot)
    
    def copy(self):
        """
        Returns a deep copy of a REBOUND simulation. You need to reset 
        any function pointers on the copy. 
        
        Returns
        ------- 
        A rebound.Simulation object.
        
        """
        w = c_int(0)
        sim = Simulation()
        clibrebound.reb_simulation_copy_with_messages(byref(sim),byref(self),byref(w))
        for majorerror, value, message in BINARY_WARNINGS:
            if w.value & value:
                if majorerror:
                    raise RuntimeError(message)
                else:  
                    # Just a warning
                    warnings.warn(message, RuntimeWarning)
        return sim

    def start_server(self, port=1234):
        """
        Start webserver on specified port. 
        The default port is 1234.
        You can access a running server by opening a web browser
        at http://localhost:1234 or http://127.0.0.1:1234
        """
        clibrebound.reb_simulation_start_server.restype = c_int
        ret_value = clibrebound.reb_simulation_start_server(byref(self), c_int(port))
        self.process_messages()
    
    def stop_server(self, port=1234):
        """
        Stop the webserver.
        """
        ret_value = clibrebound.reb_simulation_stop_server(byref(self))
        self.process_messages()


    def widget(self, port=None, host="localhost", size=(500,500)):
        if port is not None:
            port = int(port)
            if self._server_data:
                if self._server_data.contents.port != port:
                    raise RuntimeError("Server is running already on port %d but port %d was requested."%(self._server_data.contents.port, port))
        if not self._server_data:
            if port is not None:
                self.start_server(port=port)
            else:
                self.start_server()
        if not self._server_data:
            raise RuntimeError("Server did not start up immediately. Try again in a few seconds.")
        port = int(self._server_data.contents.port)
        from IPython.display import IFrame
        width, height = size
        display(IFrame("http://"+host+":"+"%d"%port, width, height))

    def cite(self):
        """
        Generate citations

        This function generates citations to papers relevant to the current 
        setting of the simulation.
        """

        txt, bib = cite(self)
        # one could check for REBOUNDx here, then append txt and bib accordingly
        print(txt + "\n\n\n" + bib)

    @property
    def simulationarchive_filename(self):
        """
        Returns the current Simulationarchive filename in use. 
        Do not set manually. Use sim.save_to_file() instead
        """
        return self._simulationarchive_filename

# Message and memory management functions
    def process_messages(self):
        clibrebound.reb_simulation_get_next_message.restype = c_int
        buf = create_string_buffer(c_int.in_dll(clibrebound, "reb_max_messages_length").value)
        while clibrebound.reb_simulation_get_next_message(byref(self), buf):
            msg = buf.value.decode("ascii")
            if msg[0]=='w':
                warnings.warn(msg[1:], RuntimeWarning)
            elif msg[0]=='e':
                raise RuntimeError(msg[1:])

# Pickling methods: return Simulationarchive binary
    def __reduce__(self):
        buf = c_char_p()
        size = c_size_t()
        clibrebound.reb_simulation_save_to_stream(byref(self), byref(buf), byref(size))
        s = bytes(string_at(buf, size=size.value)) #make copy
        clibrebound.reb_simulation_output_free_stream(buf) # free original
        return (Simulation, (s,))

# Other operators

    def __del__(self):
        if self._b_needsfree_ == 1: # to avoid, e.g., sim.particles[1]._sim.contents.G creating a Simulation instance to get G, and then freeing the C simulation when it immediately goes out of scope
            clibrebound.reb_simulation_free_pointers(byref(self))

    def __eq__(self, other):
        # This ignores the walltime parameter
        if not isinstance(other,Simulation):
            return NotImplemented
        clibrebound.reb_simulation_diff.restype = c_int
        ret = clibrebound.reb_simulation_diff(byref(self), byref(other),c_int(2))
        return not ret
            
    def diff(self, other):
        if not isinstance(other,Simulation):
            return NotImplemented
        clibrebound.reb_simulation_diff_char.restype = allocated_c_char_p
        output = clibrebound.reb_simulation_diff_char(byref(other), byref(self))
        print(output.value.decode("utf-8"))
        clibrebound.reb_free(output)

    def __add__(self, other):
        if not isinstance(other,Simulation):
            return NotImplemented
        c = self.copy()
        return c.__iadd__(other)
    
    def __iadd__(self, other):
        if not isinstance(other,Simulation):
            return NotImplemented
        clibrebound.reb_simulation_iadd.restype = c_int
        ret = clibrebound.reb_simulation_iadd(byref(self), byref(other))
        if ret==-1:
            raise RuntimeError("Cannot add simulations. Check that the simulations have the same number of particles")
        return self
    
    def __sub__(self, other):
        if not isinstance(other,Simulation):
            return NotImplemented
        c = self.copy()
        return c.__isub__(other)

    def __isub__(self, other):
        if not isinstance(other,Simulation):
            return NotImplemented
        clibrebound.reb_simulation_isub.restype = c_int
        ret = clibrebound.reb_simulation_isub(byref(self), byref(other))
        if ret==-1:
            raise RuntimeError("Cannot subtract simulations. Check that the simulations have the same number of particles")
        return self
    
    def __mul__(self, other):
        try:
            other = float(other)
        except:
            return NotImplemented
        c = self.copy()
        c.multiply(other, other)
        return c
    
    def __imul__(self, other):
        try:
            other = float(other)
        except:
            return NotImplemented
        self.multiply(other, other)
        return self

    def __rmul__(self, other):
        try:
            other = float(other)
        except:
            return NotImplemented
        c = self.copy()
        c.multiply(other, other)
        return c
    
    def __div__(self, other):
        return self.__truediv__(other)
    
    def __idiv__(self, other):
        return self.__itruediv__(other)

    def __truediv__(self, other):
        try:
            other = float(other)
        except:
            return NotImplemented
        c = self.copy()
        if other==0.:
            raise ZeroDivisionError
        c.multiply(1./other, 1./other)
        return c

    def __itruediv__(self, other):
        try:
            other = float(other)
        except:
            return NotImplemented
        if other==0.:
            raise ZeroDivisionError
        self.multiply(1./other, 1./other)
        return self

    def multiply(self, scalar_pos, scalar_vel):
        try:
            scalar_pos = float(scalar_pos)
            scalar_vel = float(scalar_vel)
        except:
            raise ValueError("Cannot multiply simulation with non-scalars.")
        clibrebound.reb_simulation_imul(byref(self), c_double(scalar_pos), c_double(scalar_vel))
    
    def rotate(self, q):
        from .rotation import Rotation
        if not isinstance(q, Rotation):
            return NotImplemented
        clibrebound.reb_simulation_irotate(byref(self), q)

#ODE functions
    def create_ode(self, length, needs_nbody=True):
        clibrebound.reb_ode_create.restype = POINTER(ODE)
        ode_p = clibrebound.reb_ode_create(byref(self), c_int(length))
        ode_p.contents.needs_nbody = c_uint(needs_nbody)
        return ODE.from_address(addressof(ode_p.contents))

# Status functions
    def status(self, showParticles=True, showAllFields=True):
        """ 
        Prints a summary of the current status 
        of the simulation.
        """
        from rebound import __version__, __build__
        s= ""
        s += "---------------------------------\n"
        s += "REBOUND version:     \t%s\n" %__version__
        s += "REBOUND built on:    \t%s\n" %__build__
        s += "Number of particles: \t%d\n" %self.N       
        s += "Selected integrator: \t" + self.integrator + "\n"       
        s += "Simulation time:     \t%.16e\n" %self.t
        s += "Current timestep:    \t%f\n" %self.dt
        s += "---------------------------------\n"
        if self.N>0 and showParticles:
            for p in self.particles:
                s += str(p) + "\n"
            s += "---------------------------------\n"
        print(s, end="")
        if showAllFields:
            print("The following fields have non-default values:")
            newsim = Simulation()
            clibrebound.reb_simulation_diff_char.restype = c_char_p
            output = clibrebound.reb_simulation_diff_char(byref(newsim), byref(self))
            print(output.decode("utf-8"))



# Set function pointer for additional forces
    @property
    def additional_forces(self):
        """
        Get or set a function pointer for calculating additional forces in the simulation.

        The argument can be a python function or something that can 
        be cast to a C function of type CFUNCTYPE(None,POINTER(Simulaton)). 
        If the forces are velocity dependent, the flag 
        force_is_velocity_dependent needs to be set to 1. Otherwise,
        the particle structures might contain incorrect velocity 
        values.
        """
        raise AttributeError("You can only set C function pointers from python.")
    @additional_forces.setter
    def additional_forces(self, func):
        self._afp = AFF(func)
        self._additional_forces = self._afp

    @property
    def pre_timestep_modifications(self):
        """
        Get or set a function pointer for pre-timestep modifications.

        The argument can be a python function or something that can be cast to a C function or a
        python function.
        """
        raise AttributeError("You can only set C function pointers from python.")
    @pre_timestep_modifications.setter
    def pre_timestep_modifications(self, func):
        self._pretmp = AFF(func)
        self._pre_timestep_modifications = self._pretmp
    
    @property
    def post_timestep_modifications(self):
        """
        Get or set a function pointer for post-timestep modifications.

        The argument can be a python function or something that can be cast to a C function or a
        python function.
        """
        raise AttributeError("You can only set C function pointers from python.")
    @post_timestep_modifications.setter
    def post_timestep_modifications(self, func):
        self._posttmp = AFF(func)
        self._post_timestep_modifications = self._posttmp
 
    @property
    def heartbeat(self):
        """
        Set a function pointer for a heartbeat function.
        The heartbeat function is called every timestep and can be used
        to monitor long simulations, check for stalled simulations and 
        output debugging information.
     
        The argument can be a python function or something that can be 
        cast to a C function or a python function.

        The function called will receive a pointer to the simulation
        object as its argument.
        
        Examples
        --------

        >>> import rebound
        >>> sim = rebound.Simulation()
        >>> sim.add(m=1.)
        >>> sim.add(m=1e-3, a=1)
        >>> def heartbeat(sim):
        >>>     # sim is a pointer to the simulation object,
        >>>     # thus use contents to access object data.
        >>>     # See ctypes documentation for details.
        >>>     print(sim.contents.t)
        >>> sim.heartbeat = heartbeat
        >>> sim.integrate(1.)

        """
        raise AttributeError("You can only set C function pointers from python (not get).")
    @heartbeat.setter
    def heartbeat(self, func):
        self._hb = AFF(func)
        self._heartbeat = self._hb

    @property 
    def coefficient_of_restitution(self):
        """
        Get or set a function pointer that defined the coefficient of restitution.
        """
        raise AttributeError("You can only set C function pointers from python.")
    @coefficient_of_restitution.setter
    def coefficient_of_restitution(self, func):
        self._corfp = CORFF(func)
        self._coefficient_of_restitution = self._corfp
    
    @property 
    def collision_resolve(self):
        """
        Get or set a function pointer for collision resolving routine.
        
        Possible options for setting:
          1) Function pointer
          2) "merge": two colliding particles will merge) 
          3) "hardsphere": two colliding particles will bounce off using a set coefficient of restitution
        """
        raise AttributeError("You can only set C function pointers from python.")
    @collision_resolve.setter
    def collision_resolve(self, func):
        if func == "merge":
            clibrebound.reb_simulation_set_collision_resolve.restype = None
            clibrebound.reb_simulation_set_collision_resolve(byref(self), clibrebound.reb_collision_resolve_merge)
        elif func == "hardsphere":
            clibrebound.reb_simulation_set_collision_resolve.restype = None
            clibrebound.reb_simulation_set_collision_resolve(byref(self), clibrebound.reb_collision_resolve_hardsphere)
        elif func == "halt":
            clibrebound.reb_simulation_set_collision_resolve.restype = None
            clibrebound.reb_simulation_set_collision_resolve(byref(self), clibrebound.reb_collision_resolve_halt)
        else:
            self._colrfp = COLRFF(func)
            self._collision_resolve = self._colrfp
    
    @property 
    def free_particle_ap(self):
        """
        Get or set a function pointer for freeing the ap pointer whenever sim.remove is called.
        """
        raise AttributeError("You can only set C function pointers from python.")
    @free_particle_ap.setter
    def free_particle_ap(self, func):
        self._fpa = FPA(func)
        self._free_particle_ap = self._fpa

# Setter/getter of parameters and constants
    @property 
    def N_real(self):
        """
        Get the current number of real (i.e. no variational/shadow) particles in the simulation.
        """
        return self.N-self.N_var

    @property
    def integrator(self):
        """
        Get or set the integrator module.

        Available integrators include:

        - ``'IAS15'`` (default)
        - ``'WHFast'``
        - ``'SEI'``
        - ``'LEAPFROG'``
        - ``'JANUS'``
        - ``'MERCURIUS'``
        - ``'WHCKL'`` 
        - ``'WHCKM'`` 
        - ``'WHCKC'`` 
        - ``'SABA4'`` 
        - ``'SABACL4'`` 
        - ``'SABACM4'`` 
        - ``'SABA(10,6,4)'`` 
        - ``'EOS'`` 
        - ``'BS'`` 
        - ``'WHFast512'``
        - ``'TRACE'``
        - ``'none'``
        
        Check the online documentation for a full description of each of the integrators. 
        """
        i = self._integrator
        for name, _i in INTEGRATORS.items():
            if i==_i:
                return name
        return i
    @integrator.setter
    def integrator(self, value):
        if isinstance(value, int):
            self._integrator = c_int(value)
        elif isinstance(value, basestring):
            value = value.lower()
            if value in INTEGRATORS: 
                self._integrator = INTEGRATORS[value]
            # Shortcuts
            elif value=="wh":
                self.integrator = "whfast"
                self.ri_whfast.corrector = 0
                self.ri_whfast.kernel = "default"
            elif value=="whc":
                self.integrator = "whfast"
                self.ri_whfast.corrector = 17
                self.ri_whfast.kernel = "default"
            elif value=="whckl":
                self.integrator = "whfast"
                self.ri_whfast.corrector = 17
                self.ri_whfast.kernel = "lazy"
            elif value=="whckm":
                self.integrator = "whfast"
                self.ri_whfast.corrector = 17
                self.ri_whfast.kernel = "modifiedkick"
            elif value=="whckc":
                self.integrator = "whfast"
                self.ri_whfast.corrector = 17
                self.ri_whfast.kernel = "composition"
            elif value[0:4]=="saba" and len(value)>4:
                self.integrator = "saba"
                self.ri_saba.type = value[4:]
            else:
                raise ValueError("Integrator not found.")
    
    @property
    def boundary(self):
        """
        Get or set the boundary module.

        Available boundary modules are:

        - ``'none'`` (default)
        - ``'open'``
        - ``'periodic'``
        - ``'shear'``
        
        Check the online documentation for a full description of each of the modules. 
        """
        i = self._boundary
        for name, _i in BOUNDARIES.items():
            if i==_i:
                return name
        return i
    @boundary.setter
    def boundary(self, value):
        if isinstance(value, int):
            self._boundary = c_int(value)
        elif isinstance(value, basestring):
            value = value.lower()
            if value in BOUNDARIES: 
                self._boundary = BOUNDARIES[value]
            else:
                raise ValueError("Warning. Boundary condition module not found.")

    @property
    def gravity(self):
        """
        Get or set the gravity module.

        Available gravity modules are:

        - ``'none'`` 
        - ``'basic'`` (default)
        - ``'compensated'``
        - ``'tree'``
        
        Check the online documentation for a full description of each of the modules. 
        """
        i = self._gravity
        for name, _i in GRAVITIES.items():
            if i==_i:
                return name
        return i
    @gravity.setter
    def gravity(self, value):
        if isinstance(value, int):
            self._gravity = c_int(value)
        elif isinstance(value, basestring):
            value = value.lower()
            if value in GRAVITIES: 
                self._gravity = GRAVITIES[value]
            else:
                raise ValueError("Warning. Gravity module not found.")

    @property
    def collision(self):
        """
        Get or set the collision module.

        Available collision modules are:

        - ``'none'`` (default)
        - ``'direct'``
        - ``'tree'``
        - ``'line'`` 
        - ``'linetree'``
        
        Check the online documentation for a full description of each of the modules. 
        """
        i = self._collision
        for name, _i in COLLISIONS.items():
            if i==_i:
                return name
        return i
    @collision.setter
    def collision(self, value):
        if isinstance(value, int):
            self._collision = c_int(value)
        elif isinstance(value, basestring):
            value = value.lower()
            if value in COLLISIONS: 
                self.collision = COLLISIONS[value]
            else:
                raise ValueError("Warning. Collision module not found.")

# Units

    @property
    def units(self):
        """
        Tuple of the units for length, time and mass.  Can be set in any order, and strings are not case-sensitive.  See ipython_examples/Units.ipynb for more information.  You can check the units' exact values and add Additional units in rebound/rebound/units.py.  Units should be set before adding particles to the simulation (will give error otherwise).

        Currently supported Units 
        -------------------------

        Times:
        Hr          : Hours
        Yr          : Julian years
        Jyr         : Julian years
        Sidereal_yr : Sidereal year
        Yr2pi       : Year divided by 2pi, with year defined as orbital period of planet at 1AU around 1Msun star
        Kyr         : Kiloyears (Julian)
        Myr         : Megayears (Julian)
        Gyr         : Gigayears (Julian)

        Lengths:
        M           : Meters
        Cm          : Centimeters
        Km          : Kilometers
        AU          : Astronomical Units

        Masses:
        Kg          : Kilograms
        Msun        : Solar masses
        Mmercury    : Mercury masses
        Mvenus      : Venus masses
        Mearth      : Earth masses
        Mmars       : Mars masses
        Mjupiter    : Jupiter masses
        Msaturn     : Saturn masses
        Muranus     : Neptune masses
        Mpluto      : Pluto masses
        
        Examples
        --------
        
        >>> sim = rebound.Simulation()
        >>> sim.units = ('yr', 'AU', 'Msun')

        """
        
        return {'length':hash_to_unit(self.python_unit_l), 'mass':hash_to_unit(self.python_unit_m), 'time':hash_to_unit(self.python_unit_t)}

    @units.setter
    def units(self, newunits):
        newunits = check_units(newunits)        
        if self.N>0: # some particles are loaded
            raise AttributeError("Error:  You cannot set the units after populating the particles array.  See ipython_examples/Units.ipynb.")
        self.update_units(newunits) 

    def update_units(self, newunits): 
        clibrebound.reb_hash.restype = c_uint32
        self.python_unit_l = clibrebound.reb_hash(c_char_p(newunits[0].encode("ascii")))
        self.python_unit_t = clibrebound.reb_hash(c_char_p(newunits[1].encode("ascii")))
        self.python_unit_m = clibrebound.reb_hash(c_char_p(newunits[2].encode("ascii")))
        self.G = convert_G(newunits)

    def equal_units(self, sim2):
        """
        Compares of this simulation to another simulation.
            
        Returns
        -------
        True if both simulations use the same units. False otherwise.
        """
        if not isinstance(sim2,Simulation):
            raise AttributeError("Error: Argument is not a Simulation object.")

        return self.python_unit_l == sim2.python_unit_l and self.python_unit_m == sim2.python_unit_m and self.python_unit_t == sim2.python_unit_t

    def convert_particle_units(self, *args): 
        """
        Will convert the units for the simulation (i.e. convert G) as well as the particles' cartesian elements.
        Must have set sim.units ahead of calling this function, so REBOUND knows what units to convert from.

        Parameters
        ----------
        3 strings corresponding to units of time, length and mass.  Can be in any order and aren't case-sensitive.        You can add new units to rebound/rebound/units.py
        """
        if self.python_unit_l == 0 or self.python_unit_m == 0 or self.python_unit_t == 0:
            raise AttributeError("Must set sim.units before calling convert_particle_units in order to know what units to convert from.")
        new_l, new_t, new_m = check_units(args)
        for p in self.particles:
            units_convert_particle(p, hash_to_unit(self.python_unit_l), hash_to_unit(self.python_unit_t), hash_to_unit(self.python_unit_m), new_l, new_t, new_m)
        self.update_units((new_l, new_t, new_m))

# Variational Equations
    def add_variation(self,order=1,first_order=None, first_order_2=None, testparticle=-1):
        """ 
        This function adds a set of variational particles to the simulation. 

        If there are N real particles in the simulation, this functions adds N additional variational 
        particles. To see how many particles (real and variational) are in a simulation, use ``'sim.N'``. 
        To see how many variational particles are in a simulation use ``'sim.N_var'``.

        Currently Leapfrog, WHFast and IAS15 support first order variational equations. IAS15 also
        supports second order variational equations.

        Parameters
        ----------
        order : integer, optional
            By default, the function adds a set of first order variational particles to the simulation. Set this flag to 2 for second order.
        first_order : Variation, optional
            Second order variational equations depend on their corresponding first order variational equations. 
            This parameter expects the Variation object corresponding to the first order variational equations. 
        first_order_2 : Variation, optional
            Same as first_order. But allows to set two different indicies to calculated off-diagonal elements. 
            If omitted, then first_order will be used for both first order equations.
        testparticle : int, optional
            If set to a value >= 0, then only one variational particle will be added and be treated as a test particle.
            

        Returns
        -------
        Returns Variation object (a copy--you can only modify it through its particles property or vary method). 
        """
        cur_N_var_config = self.N_var_config
        if order==1:
            index = clibrebound.reb_simulation_add_variation_1st_order(byref(self),c_int(testparticle))
        elif order==2:
            if first_order is None:
                raise AttributeError("Please specify corresponding first order variational equations when initializing second order variational equations.")
            if first_order_2 is None:
                first_order_2 = first_order
            index = clibrebound.reb_simulation_add_variation_2nd_order(byref(self),c_int(testparticle),c_int(first_order.index),c_int(first_order_2.index))
        else:
            raise AttributeError("Only variational equations of first and second order are supported.")

        # Need a copy because location of original might shift if more variations added
        s = Variation.from_buffer_copy(self.var_config[cur_N_var_config])

        return s
        
# MEGNO
    def init_megno(self, seed=None):
        """
        This function initialises the chaos indicator MEGNO particles and enables their integration.

        MEGNO is short for Mean Exponential Growth of Nearby orbits. It can be used to test
        if a system is chaotic or not. In the backend, the integrator is integrating an additional set
        of particles using the variational equation. Note that variational equations are better 
        suited for this than shadow particles. MEGNO is currently only supported in the IAS15 
        and WHFast integrators.

        This function also needs to be called if you are interested in the Lyapunov exponent as it is
        calculate with the help of MEGNO. See Rein and Tamayo 2015 for details on the implementation.

        For more information on MEGNO see e.g. https://dx.doi.org/10.1051/0004-6361:20011189
        """
        if seed is None:
            clibrebound.reb_simulation_init_megno(byref(self))
        else:
            clibrebound.reb_simulation_init_megno_seed(byref(self), c_uint(seed))
    
    def megno(self):
        """
        Return the current MEGNO value.
        Note that you need to call init_megno() before the start of the simulation.
        """
        if self._calculate_megno==0:
            raise RuntimeError("MEGNO cannot be calculated. Make sure to call init_megno() after adding all particles but before integrating the simulation.")

        clibrebound.reb_simulation_megno.restype = c_double
        return clibrebound.reb_simulation_megno(byref(self))
    
    def lyapunov(self):
        """
        Return the current Lyapunov Characteristic Number (LCN).
        Note that you need to call init_megno() before the start of the simulation.
        Different definitions of the LCN exist.  Here, we're following Eq 24 of 
        Cincotta and Simo (2000): https://aas.aanda.org/articles/aas/abs/2000/20/h1686/h1686.html.
        To get a timescale (the Lyapunov timescale), take the inverse of this quantity.
        """
        if self._calculate_megno==0:
            raise RuntimeError("Lyapunov Characteristic Number cannot be calculated. Make sure to call init_megno() after adding all particles but before integrating the simulation.")

        clibrebound.reb_simulation_lyapunov.restype = c_double
        return clibrebound.reb_simulation_lyapunov(byref(self))
    
# Particle add function, used to be called particle_add() and add_particle() 
    def add(self, particle=None, **kwargs):   
        """
        Adds a particle to REBOUND. Accepts one of the following:

        1) A single Particle structure.
        2) The particle's mass and a set of cartesian coordinates: m,x,y,z,vx,vy,vz.
        3) The primary as a Particle structure, the particle's mass and a set of orbital elements: primary,m,a,anom,e,omega,inv,Omega,MEAN (see :class:`.Orbit` for the definition of orbital elements).
        4) A name of an object (uses NASA Horizons to look up coordinates)
        5) A list of particles or names.
        """
        if particle is not None:
            if isinstance(particle, Particle):
                if (self.gravity == "tree" or self.collision == "tree") and self.root_size <=0.:
                    raise ValueError("The tree code for gravity and/or collision detection has been selected. However, the simulation box has not been configured yet. You cannot add particles until the the simulation box has a finite size.")
                if particle._sim:
                    if not self.equal_units(particle._sim.contents):
                        warnings.warn("Particle is being adding from a simulation that uses different units.", RuntimeWarning)
                clibrebound.reb_simulation_add(byref(self), particle)
            elif isinstance(particle, list):
                for p in particle:
                    self.add(p, **kwargs)
            elif isinstance(particle,str):
                if self.python_unit_l == 0 or self.python_unit_m == 0 or self.python_unit_t == 0:
                    self.units = ('AU', 'yr2pi', 'Msun')
                    self.G = 1.0
                builtindatasets = ["solar system", "outer solar system"]
                if particle.lower() == "solar system":          # built in test dataset
                    data.add_solar_system(self)
                elif particle.lower() == "outer solar system":  # built in test dataset
                    data.add_outer_solar_system(self)
                else:
                    if "frame" not in kwargs:
                        if hasattr(self, 'default_plane'):
                            kwargs["plane"] = self.default_plane # allow ASSIST to set default plane
                    mass_unit = hash_to_unit(self.python_unit_m) # For manually provided masses
                    self.add(horizons.query_horizons_for_particle(mass_unit, particle, **kwargs), hash=particle)
                    units_convert_particle(self.particles[-1], 'km', 's', 'kg', hash_to_unit(self.python_unit_l), hash_to_unit(self.python_unit_t), hash_to_unit(self.python_unit_m))
            else: 
                raise ValueError("Argument passed to add() not supported.")
        else: 
            self.add(Particle(simulation=self, **kwargs))
        self.process_messages()

# Particle getter functions
    @property
    def particles(self):
        """
        Returns a Particles object that allows users to access particles like a dictionary using indices, hashes, or strings. 

        The Particles object uses pointers and thus the contents update 
        as the simulation progresses. Note that the pointers could change,
        for example when a particle is added or removed from the simulation. 
        """
        particles = Particles(self)
        return particles

    @particles.deleter
    def particles(self):
        """
        Remove all particles from the simulation
        """
        clibrebound.reb_simulation_remove_all_particles(byref(self))
        self.process_messages()

    def remove(self, index=None, hash=None, keep_sorted=True):
        """ 
        Removes a particle from the simulation.

        Parameters
        ----------
        index : int, optional
            Specify particle to remove by index.
        hash : c_uint32 or string, optional
            Specifiy particle to remove by hash (if a string is passed, the corresponding hash is calculated).
        keep_sorted : bool, optional
            By default, remove preserves the order of particles in the particles array. 
            Might set it to zero in cases with many particles and many removals to speed things up.
        """
        if index is not None:
            clibrebound.reb_simulation_remove_particle(byref(self), index, keep_sorted)
        if hash is not None:
            hash_types = c_uint32, c_uint, c_uint64
            PY3 = sys.version_info[0] == 3
            if PY3:
                string_types = str,
                int_types = int,
            else:
                string_types = basestring,
                int_types = int, long
            if isinstance(hash, string_types):
                clibrebound.reb_simulation_remove_particle_by_hash(byref(self), rebhash(hash), keep_sorted)
            elif isinstance(hash, int_types):
                clibrebound.reb_simulation_remove_particle_by_hash(byref(self), c_uint32(hash), keep_sorted)
            elif isinstance(hash, hash_types):
                clibrebound.reb_simulation_remove_particle_by_hash(byref(self), hash, keep_sorted)

        self.process_messages()

# Orbit calculation
    def orbits(self, primary=None, jacobi_masses=False):
        """ 
        Calculate orbital parameters for all particles in the simulation.
        By default, this functions returns the orbits in Jacobi coordinates. 

        If MEGNO is enabled, variational particles will be ignored.

        Parameters
        ----------

        primary     : rebound.Particle, optional
            Set the primary against which to reference the osculating orbit. Default (use Jacobi center of mass).
            For heliocentric coordinates, pass the central object to this parameter. 
        jacobi_masses: bool
            Whether to use jacobi primary mass in orbit calculation. (Default: False)

        Returns
        -------
        Returns an array of Orbits of length N-1.
        """
        orbits = []
       
        if primary is None:
            jacobi = True
            primary = self.particles[0]
            clibrebound.reb_particle_com_of_pair.restype = Particle
        else:
            jacobi = False

        for p in self.particles[1:self.N_real]:
            if jacobi_masses is True:
                interior_mass = primary.m
                # orbit conversion uses mu=G*(p.m+primary.m) so set prim.m=Mjac-m so mu=G*Mjac
                primary.m = self.particles[0].m*(p.m + interior_mass)/interior_mass - p.m
                orbits.append(p.orbit(primary=primary))
                primary.m = interior_mass # back to total mass of interior bodies to update com
            else:
                orbits.append(p.orbit(primary=primary))
            if jacobi is True: # update com to include current particle for next iteration
                primary = clibrebound.reb_particle_com_of_pair(primary, p)

        return orbits

# COM calculation 
    def com(self, first=0, last=None):
        """
        Returns the center of momentum for all particles in the simulation.

        Parameters
        ----------
        first: int, optional
            If ``first`` is specified, only calculate the center of momentum starting
            from index=``first``.
        last : int or None, optional
            If ``last`` is specified only calculate the center of momentum up to 
            (but excluding) index=``last``.  Same behavior as Python's range function.

        Examples
        --------
        >>> sim = rebound.Simulation()
        >>> sim.add(m=1, x=-20)
        >>> sim.add(m=1, x=-10)
        >>> sim.add(m=1, x=0)
        >>> sim.add(m=1, x=10)
        >>> sim.add(m=1, x=20)
        >>> com = sim.com()
        >>> com.x
        0.0 
        >>> com = sim.com(first=2,last=4) # Considers indices 2,3
        >>> com.x
        5.0

        """
        if last is None:
            last = self.N_real
        clibrebound.reb_simulation_com_range.restype = Particle
        return clibrebound.reb_simulation_com_range(byref(self), c_int(first), c_int(last))

# Tools
    def serialize_particle_data(self,**kwargs):
        """
        Fast way to access serialized particle data via numpy arrays.

        This function can directly set the values of numpy arrays to
        current particle data. This is significantly faster than accessing
        particle data via `sim.particles` as all the copying is done 
        on the C side. 
        No memory is allocated by this function.
        It expects correctly sized numpy arrays as arguments. The argument
        name indicates what kind of particle data is written to the array. 
        
        Possible argument names are "hash", "m", "r", "xyz", "vxvyvz", and 
        "xyzvxvyvz". The datatype for the "hash" array needs to be uint32. 
        The other arrays expect a datatype of float64. The lengths of 
        "hash", "m", "r" arrays need to be at least sim.N. The lengths of 
        xyz and vxvyvz need to be at least 3*sim.N. The length of
        "xyzvxvyvz" arrays need to be 6*sim.N. Exceptions are raised 
        otherwise.

        Note that this routine is only intended for special use cases
        where speed is an issue. For normal use, it is recommended to
        access particle data via the `sim.particles` array. Be aware of
        potential issues that arise by directly accesing the memory of
        numpy arrays (see numpy documentation for more details).

        Examples
        --------
        This sets an array to the xyz positions of all particles:

        >>> import numpy as np
        >>> a = np.zeros((sim.N,3),dtype="float64")
        >>> sim.serialize_particle_data(xyz=a)
        >>> print(a)

        To get all current radii of particles:

        >>> a = np.zeros(sim.N,dtype="float64")
        >>> sim.serialize_particle_data(r=a)
        >>> print(a)
        
        To get all current radii and hashes of particles:

        >>> a = np.zeros(sim.N,dtype="float64")
        >>> b = np.zeros(sim.N,dtype="uint32")
        >>> sim.serialize_particle_data(r=a,hash=b)
        >>> print(a,b)

        """
        N = self.N
        possible_keys = ["hash","m","r","xyz","vxvyvz","xyzvxvyvz"]
        d = {x:None for x in possible_keys}
        for k,v in kwargs.items():
            if k in d:
                if k == "hash":
                    if v.dtype!= "uint32":
                        raise AttributeError("Expected 'uint32' data type for '%s' array."%k)
                    if v.size<N:
                        raise AttributeError("Array '%s' is not large enough."%k)
                    d[k] = v.ctypes.data_as(POINTER(c_uint32))
                else:
                    if v.dtype!= "float64":
                        raise AttributeError("Expected 'float64' data type for %s array."%k)
                    if k in ["xyz", "vxvyvz"]:
                        minsize = 3*N
                    elif k in ["xyzvxvyvz"]:
                        minsize = 6*N
                    else:
                        minsize = N
                    if v.size<minsize:
                        raise AttributeError("Array '%s' is not large enough."%k)
                    d[k] = v.ctypes.data_as(POINTER(c_double))
            else:
                raise AttributeError("Only '%s' are currently supported attributes for serialization." % "', '".join(d.keys()))

        clibrebound.reb_simulation_get_serialized_particle_data(byref(self), d["hash"], d["m"], d["r"], d["xyz"], d["vxvyvz"], d["xyzvxvyvz"])
    
    def set_serialized_particle_data(self,**kwargs):
        """
        Fast way to set serialized particle data via numpy arrays.
        This is the inverse of Simulation.serialize_particle_data()
        and uses the same syntax
        """

        N = self.N
        possible_keys = ["hash","m","r","xyz","vxvyvz","xyzvxvyvz"]
        d = {x:None for x in possible_keys}
        for k,v in kwargs.items():
            if k in d:
                if k == "hash":
                    if v.dtype!= "uint32":
                        raise AttributeError("Expected 'uint32' data type for '%s' array."%k)
                    if v.size<N:
                        raise AttributeError("Array '%s' is not large enough."%k)
                    d[k] = v.ctypes.data_as(POINTER(c_uint32))
                else:
                    if v.dtype!= "float64":
                        raise AttributeError("Expected 'float64' data type for %s array."%k)
                    if k in ["xyz", "vxvyvz"]:
                        minsize = 3*N
                    elif k in ["xyzvxvyvz"]:
                        minsize = 6*N
                    else:
                        minsize = N
                    if v.size<minsize:
                        raise AttributeError("Array '%s' is not large enough."%k)
                    d[k] = v.ctypes.data_as(POINTER(c_double))
            else:
                raise AttributeError("Only '%s' are currently supported attributes for serialization." % "', '".join(d.keys()))

        clibrebound.reb_simulation_set_serialized_particle_data(byref(self), d["hash"], d["m"], d["r"], d["xyz"], d["vxvyvz"], d["xyzvxvyvz"])

    def move_to_hel(self):
        """
        This function moves all particles in the simulation to the heliocentric frame.
        Note that the simulation will not stay in the heliocentric frame during integrations
        as the heliocentric frame is not an inertial frame.
        """
        clibrebound.reb_simulation_move_to_hel(byref(self))
    
    def move_to_com(self):
        """
        This function moves all particles in the simulation to a center of momentum frame.
        In that frame, the center of mass is at the origin and does not move.
        It makes sense to call this function at the beginning of the integration, especially 
        for the high accuracy integrators IAS15 and WHFast.
        """
        clibrebound.reb_simulation_move_to_com(byref(self))

    def energy(self):
        """
        Returns the sum of potential and kinetic energy of all particles in the simulation.
        """
        clibrebound.reb_simulation_energy.restype = c_double
        return clibrebound.reb_simulation_energy(byref(self))
   
    def angular_momentum(self):
        """
        Returns a list of the three (x,y,z) components of the total angular momentum of all particles in the simulation.
        """
        clibrebound.reb_simulation_angular_momentum.restype = Vec3dBasic
        return Vec3d(clibrebound.reb_simulation_angular_momentum(byref(self)))
    
    def ias15_timescale(self):
        """
        Returns the gravitational timescale in code units. See Pham, Rein, Spiegel 2023 for details. 
        """
        clibrebound.reb_integrator_ias15_timescale.restype = c_double
        return clibrebound.reb_integrator_ias15_timescale(byref(self))

    def configure_box(self, boxsize, N_root_x=1, N_root_y=1, N_root_z=1):
        """
        Initialize the simulation box.

        This function only needs to be called it boundary conditions other than "none"
        are used. In such a case the boxsize must be known and is set with this function.

        Parameters
        ----------
        boxsize : float
            The size of one root box.
        N_root_x, N_root_y, N_root_z : int, optional
            The number of root boxes in each direction. The total size of the simulation box
            will be ``N_root_x * boxsize``, ``N_root_y * boxsize`` and ``N_root_z * boxsize``.
            By default, there will be exactly one root box in each direction.
        """
        clibrebound.reb_simulation_configure_box(byref(self), c_double(boxsize), c_int(N_root_x), c_int(N_root_y), c_int(N_root_z))
        return
   
# Output to file (Simulationarchive)
    def save_to_file(self, filename, interval=None, walltime=None, step=None, delete_file=False):
        """
        Saves a simulation to a file (Simulationarchive binary format). 

        If just the filename is passed, then the simulation is saved immediately. 
        You can also automate taking snapshots during a simulation. For that specify
        the interval at which outputs are required. You can do that using either
        `interval` (physical time), `walltime` (in seconds), or `step` 
        (the number of timesteps in-between snapshots). 

        
        Arguments
        ---------
        filename : str
            Filename of the binary file.
        interval : float
            Interval between outputs in code units.
        walltime : float
            Interval between outputs in wall time (seconds). 
            Useful when using IAS15 with adaptive timesteps. 
        step : int
            Interval between outputs in number of timesteps. 
            Useful when outputs need to be spaced exactly.
        delete_file: bool
            False (default) appends to file if it exists.
            True deletes the file first.
        
        Examples
        --------
        The following example shows how to output a simulation and
        then read it back in. 

        >>> sim = rebound.Simulation()
        >>> sim.add(m=1.)
        >>> sim.save_to_file("test.bin")
        >>> sim2 = rebound.Simulation("test.bin")

        The following example creates a simulation, then 
        initializes the Simulationarchive and integrates
        it forward in time. 

        >>> sim = rebound.Simulation()
        >>> sim.add(m=1.)
        >>> sim.add(m=1.e-3,x=1.,vy=1.)
        >>> sim.save_to_file("sa.bin",interval=1000.)
        >>> sim.integrate(1e8)

        The Simulationarchive can later be read in using the following syntax:

        >>> sa = rebound.Simulationarchive("sa.bin")
        >>> sim = sa[0]   # get the first snapshot in the SA file (initial conditions)
        >>> sim = sa[-1]  # get the last snapshot in the SA file

        """
        if delete_file and os.path.isfile(filename):
            os.remove(filename)

        modes = sum(1 for i in [interval, walltime,step] if i is not None)

        if modes == 0:
            # Immediately save to file.
            clibrebound.reb_simulation_save_to_file(byref(self), c_char_p(filename.encode("ascii")))
        elif modes == 1:
            if delete_file:
                # reset intervals so that automate functions in C set sim->next consistently
                self.simulationarchive_auto_interval = 0
                self.simulationarchive_auto_walltime = 0
                self.simulationarchive_auto_step = 0
            if interval:
                clibrebound.reb_simulation_save_to_file_interval(byref(self), c_char_p(filename.encode("ascii")), c_double(interval))
            if walltime:
                clibrebound.reb_simulation_save_to_file_walltime(byref(self), c_char_p(filename.encode("ascii")), c_double(walltime))
            if step:
                clibrebound.reb_simulation_save_to_file_step(byref(self), c_char_p(filename.encode("ascii")), c_uint64(step))
            self.process_messages()
        else:
            raise AttributeError("Cannot specify more than one of interval, walltime, or step.")

    def output_screenshot(self, filename):
        """
        Saves a screenshot of the current WebGL visualization to a file in png format.

        The web server needs to be started, and a web browser needs to be 
        connected to the server in order to take screen shots.

        Arguments
        ---------
        filename : str
            Filename of the output file.
        
        Examples
        --------
        The following example take a screenshot of a simulation.

        >>> sim = rebound.Simulation()
        >>> sim.integrator = "whfast"
        >>> sim.add(m=1.)
        >>> sim.add(a=1.)
        >>> sim.widget() # this connects one client
        >>> sim.output_screenshot("screenshot.png")
        """
                
        clibrebound.reb_simulation_output_screenshot(byref(self), c_char_p(filename.encode("ascii")))
        self.process_messages()

# Integration
    def step(self):
        """
        Perform exactly one integration step with REBOUND. This function is rarely needed.
        Instead, use integrate().
        """
        clibrebound.reb_simulation_step(byref(self))
        self.process_messages()
    
    def steps(self, N_steps):
        """
        Perform exactly N_steps integration steps with REBOUND. This function is rarely needed.
        Instead, use integrate().
        """
        clibrebound.reb_simulation_steps(byref(self),c_uint(N_steps))
        self.process_messages()

    def integrate(self, tmax, exact_finish_time=1):
        """
        Main integration function. Call this function when you have setup your simulation and want to integrate it forward (or backward) in time. The function might be called many times to integrate the simulation in steps and create outputs in-between steps.
        
        Parameters
        ----------
        tmax : float
            The final time of your simulation. If the current time is 100, and tmax=200, then after the calling the integrate routine, the time has advanced to t=200. If tmax is larger than or equal to the current time, no integration will be performed.
        exact_finish_time: int, optional
            This argument determines whether REBOUND should try to finish at the exact time (tmax) you give it or if it is allowed to overshoot. Overshooting could happen if one starts at t=0, has a timestep of dt=10 and wants to integrate to tmax=25. With ``exact_finish_time=1``, the integrator will choose the last timestep such that t is exactly 25 after the integration, otherwise t=30. Note that changing the timestep does affect the accuracy of symplectic integrators negatively.
        
        Exceptions
        ----------
        Exceptions are thrown when no more particles are left in the simulation or when a generic integration error occured. 
        If you specified exit_min_distance or exit_max_distance, then additional exceptions might thrown for escaping particles or particles that undergo a clos encounter.
        
        Examples
        -------- 
        The typical usage is as follows. Note the use of ``np.linspace`` to create equally spaced outputs. 
        Using ``np.logspace`` can be used to easily produce logarithmically spaced outputs.

        >>> import numpy as np
        >>> for time in np.linspace(0,100.,10): 
        >>>     sim.integrate(time)
        >>>     perform_output(sim)
        
        """
        self.exact_finish_time = c_int(exact_finish_time)
        ret_value = clibrebound.reb_simulation_integrate(byref(self), c_double(tmax))
        if ret_value == 1:
            self.process_messages()
            raise GenericError("An error occured during the integration.")
        if ret_value == 2:
            if self._N_odes>0:
                raise NoParticles("No particles found. Will exit. Use BS integrator to integrate user-defined ODEs without any particles present.");
            else:
                raise NoParticles("No more particles left in simulation.")
        if ret_value == 3:
            raise Encounter("Two particles had a close encounter (d<exit_min_distance).")
        if ret_value == 4:
            raise Escape("A particle escaped (r>exit_max_distance).")
        if ret_value == 5:
            pass # User caused exit. Do not raise error message
        if ret_value == 6:
            raise KeyboardInterrupt
        if ret_value == 7:
            raise Collision("Two particles collided (d < r1+r2)")
        self.process_messages()

    def stop(self):
        """
        Call this function to stop an integration, for example
        from the heartbeat function.
        """
        clibrebound.reb_simulation_stop(byref(self))

    def reset_integrator(self):
        """
        Call this function to reset temporary integrator variables
        """
        clibrebound.reb_simulation_reset_integrator(byref(self))

    def synchronize(self):
        """
        Call this function if safe-mode is disabled and you need to synchronize particle positions and velocities between timesteps.
        """
        clibrebound.reb_simulation_synchronize(byref(self))
        self.process_messages()
    
    def update_tree(self):
        """
        Call this function to update the tree structure manually after removing particles.
        """
        clibrebound.reb_simulation_update_tree(byref(self))
    

class timeval(Structure):
    _fields_ = [("tv_sec",c_int64),("tv_usec",c_int64)]

from .particle import Particle
from .particles import Particles


from .integrators.bs import ODE, IntegratorBS
from .integrators.whfast import IntegratorWHFast
from .integrators.whfast512 import IntegratorWHFast512
from .integrators.janus import IntegratorJanus
from .integrators.sei import IntegratorSEI
from .integrators.eos import IntegratorEOS
from .integrators.ias15 import IntegratorIAS15
from .integrators.saba import IntegratorSABA
from .integrators.mercurius import IntegratorMercurius
from .integrators.trace import IntegratorTRACE

from .variation import Variation

# Setting up fields after class definition (because of self-reference)

class ServerData(Structure):
    _fields_ = [
            ("r", POINTER(Simulation)),
            ("_screenshot", c_void_p),
            ("_N_screenshot", c_size_t),
            ("_status_before_screenshot", c_int),
            ("port", c_int),
            ("need_copy", c_int),
            ("ready", c_int),
            # other fields not needed.
            ]

Simulation._fields_ = [
                ("t", c_double),
                ("G", c_double),
                ("softening", c_double),
                ("dt", c_double),
                ("dt_last_done", c_double),
                ("steps_done", c_uint64),
                ("N", c_uint),
                ("N_var", c_int),
                ("N_var_config", c_int),
                ("var_config", POINTER(Variation)),
                ("_var_rescale_warning", c_int),
                ("N_active", c_int),
                ("testparticle_type", c_int),
                ("testparticle_hidewarnings", c_int),
                ("_particle_lookup_table", POINTER(HashPointerPair)),
                ("hash_ctr", c_int),
                ("N_lookup", c_int),
                ("N_allocated_lookup", c_int),
                ("N_allocated", c_uint),
                ("_particles", POINTER(Particle)),
                ("gravity_cs", POINTER(Vec3dBasic)),
                ("N_allocated_gravity_cs", c_int),
                ("_tree_root", c_void_p),
                ("_tree_needs_update", c_int),
                ("opening_angle2", c_double),
                ("_status", c_int),
                ("exact_finish_time", c_int),
                ("force_is_velocity_dependent", c_uint),
                ("gravity_ignore", c_uint),
                ("_output_timing_last", c_double),
                ("save_messages", c_int),
                ("messages", c_void_p),
                ("exit_max_distance", c_double),
                ("exit_min_distance", c_double),
                ("usleep", c_double),
                ("_display_view", c_void_p),
                ("_display_data", c_void_p), # not needed from python
                ("_server_data", POINTER(ServerData)),
                ("track_energy_offset", c_int),
                ("energy_offset", c_double),
                ("walltime", c_double),
                ("walltime_last_step", c_double),
                ("walltime_last_steps", c_double),
                ("_walltime_last_steps_sum", c_double),
                ("_walltime_last_steps_N", c_int),
                ("python_unit_t",c_uint32),
                ("python_unit_l",c_uint32),
                ("python_unit_m",c_uint32),
                ("boxsize", Vec3dBasic),
                ("boxsize_max", c_double),
                ("root_size", c_double),
                ("N_root", c_int),
                ("N_root_x", c_int),
                ("N_root_y", c_int),
                ("N_root_z", c_int),
                ("N_ghost_x", c_int),
                ("N_ghost_y", c_int),
                ("N_ghost_z", c_int),
                ("collision_resolve_keep_sorted", c_int),
                ("collisions", c_void_p),
                ("N_allocated_collisions", c_int),
                ("collisions_N", c_uint),
                ("minimum_collision_velocity", c_double),
                ("collisions_plog", c_double),
                ("collisions_log_n", c_int64),
                ("_calculate_megno", c_int),
                ("_megno_Ys", c_double),
                ("_megno_Yss", c_double),
                ("_megno_cov_Yt", c_double),
                ("_megno_var_t", c_double),
                ("_megno_mean_t", c_double),
                ("_megno_mean_Y", c_double),
                ("_megno_initial_t", c_double),
                ("_megno_n", c_int64),
                ("rand_seed",c_uint),
                ("simulationarchive_version", c_int),
                ("simulationarchive_auto_interval", c_double),
                ("simulationarchive_auto_walltime", c_double),
                ("simulationarchive_auto_step", c_uint64),
                ("simulationarchive_next", c_double),
                ("simulationarchive_next_step", c_uint64),
                ("_simulationarchive_filename", c_char_p),
                ("_collision", c_int),
                ("_integrator", c_int),
                ("_boundary", c_int),
                ("_gravity", c_int),
                ("ri_sei", IntegratorSEI), 
                ("ri_whfast", IntegratorWHFast),
                ("ri_whfast512", IntegratorWHFast512),
                ("ri_saba", IntegratorSABA),
                ("ri_ias15", IntegratorIAS15),
                ("ri_mercurius", IntegratorMercurius),
                ("ri_trace", IntegratorTRACE),
                ("ri_janus", IntegratorJanus),
                ("ri_eos", IntegratorEOS),
                ("ri_bs", IntegratorBS),
                ("_odes", POINTER(POINTER(ODE))),
                ("_N_odes", c_int),
                ("_N_allocated_odes", c_int),
                ("_odes_warnings", c_int),
                ("_additional_forces", CFUNCTYPE(None,POINTER(Simulation))),
                ("_pre_timestep_modifications", CFUNCTYPE(None,POINTER(Simulation))),
                ("_post_timestep_modifications", CFUNCTYPE(None,POINTER(Simulation))),
                ("_heartbeat", CFUNCTYPE(None,POINTER(Simulation))),
                ("_key_callback", CFUNCTYPE(c_int,POINTER(Simulation), c_int)),
                ("_coefficient_of_restitution", CFUNCTYPE(c_double,POINTER(Simulation), c_double)),
                ("_collision_resolve", CFUNCTYPE(c_int,POINTER(Simulation), CollisionS)),
                ("_free_particle_ap", CFUNCTYPE(None, POINTER(Particle))),
                ("_extras_cleanup", CFUNCTYPE(None, POINTER(Simulation))),
                ("extras", c_void_p),
                 ]


AFF = CFUNCTYPE(None,POINTER(Simulation))
CORFF = CFUNCTYPE(c_double,POINTER(Simulation), c_double)
COLRFF = CFUNCTYPE(c_int, POINTER(Simulation), CollisionS)
FPA = CFUNCTYPE(None, POINTER(Particle))

# Check if Simulation has same size upon import. Nothing will work if there is a mismatch.
clibrebound.reb_simulation_struct_size.res_type = c_size_t
simulation_size_c = clibrebound.reb_simulation_struct_size()

if simulation_size_c != sizeof(Simulation):
    raise ImportError("Size of C struct `reb_simulation` differs from size of python class `Simulation`. This will lead to undefined behaviour. Make sure the C and Python versions of REBOUND are up-to-date and consistent. If you have installed REBOUND from source, you might need to recompile / reinstall the python package.")


# Import at the end to avoid circular dependence
from . import horizons
from . import data
from .simulationarchive import Simulationarchive



import ctypes
from .tools import M_to_E
from .vectors import Vec3dBasic

class Orbit(ctypes.Structure):
    """
    A class containing orbital parameters for a particle.
    This is an abstraction of the reb_orbit data structure in C.

    When using the various REBOUND functions using Orbits, all angles are in radians. 
    The following image illustrated the most important angles used.
    In REBOUND the reference direction is the positive x direction, the reference plane
    is the xy plane.
    
    .. image:: images/orbit.png
       :width: 500px
       :height: 450px

    Image from wikipedia. CC-BY-SA-3.

    Attributes
    ----------
    d       : float           
        radial distance from reference 
    v       : float         
        velocity relative to central object's velocity
    h       : float           
        specific angular momentum
    P       : float           
        orbital period (negative if hyperbolic)
    n       : float           
        mean motion    (negative if hyperbolic)
    a       : float           
        semimajor axis
    e       : float           
        eccentricity
    inc     : float           
        inclination
    Omega   : float           
        longitude of ascending node
    omega   : float           
        argument of pericenter
    pomega  : float           
        longitude of pericenter
    f       : float           
        true anomaly
    M       : float           
        mean anomaly
    E       : float           
        eccentric anomaly (requires solving Kepler's equation - only calculated when needed)
    l       : float           
        mean longitude = Omega + omega + M
    theta   : float           
        true longitude = Omega + omega + f
    T       : float
        time of pericenter passage
    rhill   : float
        Hill radius ( =a*pow(m/(3M),1./3.) )
    pal_h   : float
        Cartesian component of the eccentricity ( = e*sin(pomega) )
    pal_k   : float
        Cartesian component of the eccentricity ( = e*cos(pomega) )
    pal_ix   : float
        Cartesian component of the inclination ( = 2*sin(i/2)*cos(Omega) )
    pal_iy   : float
        Cartesian component of the inclination ( = 2*sin(i/2)*sin(Omega) )
    hvec     : Vec3d
        Specific angular momentum vector
    evec     : Vec3d
        Eccentricity vector (mag = eccentricity, points toward pericenter)
    """
    _fields_ = [("d", ctypes.c_double),
                ("v", ctypes.c_double),
                ("h", ctypes.c_double),
                ("P", ctypes.c_double),
                ("n", ctypes.c_double),
                ("a", ctypes.c_double),
                ("e", ctypes.c_double),
                ("inc", ctypes.c_double),
                ("Omega", ctypes.c_double),
                ("omega", ctypes.c_double),
                ("pomega", ctypes.c_double),
                ("f", ctypes.c_double),
                ("M", ctypes.c_double),
                ("l", ctypes.c_double),
                ("theta", ctypes.c_double),
                ("T", ctypes.c_double),
                ("rhill", ctypes.c_double),
                ("pal_h", ctypes.c_double),
                ("pal_k", ctypes.c_double),
                ("pal_ix", ctypes.c_double),
                ("pal_iy", ctypes.c_double),
                ("hvec", Vec3dBasic),
                ("evec", Vec3dBasic)]

    def __repr__(self):
        return "<rebound.Orbit instance, a={0} e={1} inc={2} Omega={3} omega={4} f={5}>".format(str(self.a),str(self.e), str(self.inc), str(self.Omega), str(self.omega), str(self.f))

    @property 
    def E(self):
        return M_to_E(self.e, self.M)




import ctypes
from ctypes import cast

from .. import clibrebound
from ..simulation import Simulation
from ..particle import Particle
from ..vectors import Vec3dBasic

TRACE_PERI_MODES = {"FULL_BS": 1, "PARTIAL_BS": 0, "FULL_IAS15": 2}

class IntegratorTRACE(ctypes.Structure):
    """
    This class is an abstraction of the C-struct reb_integrator_trace.
    It controls the behaviour of the TRACE integrator.  See Lu, Hernandez and Rein (2024)
    for more details.

    :ivar float r_crit_hill:
        Switching radius in units of the modified hill radius.
    :ivar float peri_mode:
        Pericenter switching mode. Determines how close encounters with the central star are integrated.
        The default is 1, which uses the FULL BS prescription. This converts our system from Democratic Heliocentric Coordinates to back to Intertial Cartesian coordinates. The whole system is integrated with Bulirsch-Stoer.
        If set to 0, we use the PARTIAL BS prescription. This moves all terms from the Jump step to the Kepler step as described in Hernandez and Dehnen (2023). The Kepler step is integrated with Bulirsch-Stoer.
        If set to 2, we use the FULL IAS15 prescription. This acts similarly to the PURE BS prescription, but uses IAS15 instead of BS to integrate the whole system.
    :ivar float peri_crit_eta:
        Pericenter switching condition: ratio of timestep to condition from Pham, Rein and Spiegel 2024.

    Example usage:

    >>> sim = rebound.Simulation()
    >>> sim.integrator = "trace"
    >>> sim.ri_trace.r_crit_hill = 4
    >>> sim.ri_trace.peri_crit_eta = 1

    """
    def __repr__(self):
        return '<{0}.{1} object at {2}, r_crit_hill={3}, peri_mode=={4}, peri_crit_eta=={5}>'.format(self.__module__, type(self).__name__, hex(id(self)), self.r_crit_hill, self.peri_mode, self.peri_crit_eta)

    _fields_ = [("_S", ctypes.CFUNCTYPE(ctypes.c_int, ctypes.POINTER(Simulation), ctypes.c_uint, ctypes.c_uint)),
                ("_S_peri", ctypes.CFUNCTYPE(ctypes.c_int, ctypes.POINTER(Simulation), ctypes.c_uint)),
                ("peri_mode", ctypes.c_uint),
                ("r_crit_hill", ctypes.c_double),
                ("peri_crit_eta", ctypes.c_double),
                ("_mode", ctypes.c_uint),
                ("_encounter_N", ctypes.c_uint),
                ("_encounter_N_active", ctypes.c_uint),
                ("_N_allocated", ctypes.c_uint),
                ("_N_allocated_additionalforces", ctypes.c_uint),
                ("_tponly_encounter", ctypes.c_uint),
                ("_particles_backup", ctypes.POINTER(Particle)),
                ("_particles_backup_kepler", ctypes.POINTER(Particle)),
                ("_particles_backup_additional_forces", ctypes.POINTER(Particle)),
                ("_encounter_map", ctypes.POINTER(ctypes.c_int)),
                ("_com_pos", Vec3dBasic),
                ("_com_vel", Vec3dBasic),
                ("_current_Ks", ctypes.POINTER(ctypes.c_int)),
                ("_current_C", ctypes.c_uint),
                ("_force_accept", ctypes.c_uint),
                ]
    @property
    def S(self):
        raise AttributeError("You can only set C function pointers from python.")
    @S.setter
    def S(self, func):
        if func == "default":
            self._S = cast(clibrebound.reb_integrator_trace_switch_default,TRACEKF)
        else:
            self._Sfp = TRACEKF(func)
            self._S = self._Sfp

    @property
    def S_peri(self):
        raise AttributeError("You can only set C function pointers from python.")
    @S_peri.setter
    def S_peri(self, func):
        if func == "default":
            self._S_peri = cast(clibrebound.reb_integrator_trace_switch_peri_default,TRACECF)
        elif func == "none":
            self._S_peri = cast(clibrebound.reb_integrator_trace_switch_peri_none,TRACECF)
        else:
            self._S_perifp = TRACECF(func)
            self._S_peri = self._S_perifp
    
    @property
    def peri_mode(self):
        """
        Get or set the pericenter switching mode.

        Available kernels are:

        - ``'FULL_BS'`` (Integrate entire system with BS)
        - ``'PARTIAL_BS'`` (Integrate only the Kepler step with BS)
        - ``'FULL_IAS15'`` (Integrate entire system with IAS15)
        """
        i = self._peri_mode
        for name, _i in TRACE_PERI_MODES.items():
            if i==_i:
                return name
        return i
    @peri_mode.setter
    def peri_mode(self, value):
        if isinstance(value, int):
            self._peri_mode = ctypes.c_uint(value)
        elif isinstance(value, basestring):
            if value in TRACE_PERI_MODES: 
                self._peri_mode = TRACE_PERI_MODES[value]
            else:
                raise ValueError("Warning. Pericenter switching mode not found.")

TRACEKF = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.POINTER(Simulation), ctypes.c_uint, ctypes.c_uint)
TRACECF = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.POINTER(Simulation), ctypes.c_uint)



import ctypes

class IntegratorSEI(ctypes.Structure):
    """
    This class is an abstraction of the C-struct reb_integrator_sei.
    It controls the behaviour of the symplectic SEI integrator for shearing
    sheet calculations. It is described in Rein and Tremaine (2011).
    
    This struct should be accessed via the simulation class only. Here is an 
    example:

    >>> sim = rebound.Simulation()
    >>> sim.ri_sei.OMEGA =  1.58
    
    :ivar float OMEGA:          
        The epicyclic frequency OMEGA. For simulations making use of shearing 
        sheet boundary conditions, REBOUND needs to know the epicyclic frequency. 
        By default, OMEGA is 1. For more details read Rein and Tremaine 2011.
    :ivar float OMEGAZ:          
        The z component of the epicyclic frequency OMEGA. By default, it is assuming
        OMEGAZ is the same as OMEGA.
    """
    _fields_ = [("OMEGA", ctypes.c_double),
                ("OMEGAZ", ctypes.c_double),
                ("_lastdt", ctypes.c_double),
                ("_sindt", ctypes.c_double),
                ("_tandt", ctypes.c_double),
                ("_sindtz", ctypes.c_double),
                ("_tandtz", ctypes.c_double)]




import rebound
import unittest
import numpy as np
import math
import sys
import warnings
from datetime import datetime

def chaotic_exchange_sim():
    sim = rebound.Simulation()
    # Setup using xyz instead of orbital elements for
    # machine independent test

    star_m = 1;
    jup_m = 0.01 / (star_m - 0.01);

    jup_a = 5.2;
    jup_e = 0.0;
    star_x = -(jup_m / (star_m + jup_m)) * (jup_a * (1 + jup_e));
    star_vy = -(jup_m / (star_m + jup_m)) * np.sqrt(((star_m + jup_m) / jup_a) * ((1 - jup_e) / (1 + jup_e)));

    jup_x = (star_m / (star_m + jup_m)) * (jup_a * (1 + jup_e));
    jup_vy = (star_m / (star_m + jup_m)) * np.sqrt(((star_m + jup_m) / jup_a) * ((1 - jup_e) / (1 + jup_e)));

    t_x = 4.42;
    t_vy = 0.0072 * (365.25) * (1 / (2 * np.pi))

    sim = rebound.Simulation()
    sim.add(m=star_m, x=star_x, vy=star_vy)
    sim.add(m=jup_m, x=jup_x, vy = jup_vy)
    sim.add(m=0, x=t_x + star_x, vy = t_vy + star_vy)
    return sim

def pericenter_sim():
    sim = rebound.Simulation()
    sun = rebound.Particle(m=1.)
    sim.add(sun)
    sim.add(primary=sun, m=9.55e-4, a=5.2)
    sim.add(primary=sun, m=2.857e-4, a=9.58,e=0.99,inc=math.pi/2.)
    sim.move_to_com()
    return sim

def derivatives_ho(ode, yDot, y, t):
    m = 1.
    k = 100.
    yDot[0] = y[1]
    yDot[1] = -k/m*y[0]

class TestIntegratorTraceHarmonic(unittest.TestCase):
   
    def test_trace_harmonic_with_nbody(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.123);
        sim.add(m=1e-3,a=2.6,e=0.123);
        sim.integrator = "TRACE"
        ode_ho = sim.create_ode(length=2, needs_nbody=False)
        ode_ho.derivatives = derivatives_ho

        ode_ho.y[0] = 1. 
        ode_ho.y[1] = 0. # zero velocity

        sim.integrate(20.*math.pi)
        self.assertLess(math.fabs(ode_ho.y[0]-1.),2e-10)
        self.assertLess(math.fabs(ode_ho.y[1]),2e-9)
    
    def test_trace_harmonic_with_nbody_coupledy(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1,e=0.123);
        sim.add(m=1e-3,a=2.6,e=0.123);
        sim.integrator = "TRACE"
        e0 = sim.energy()
        ode_ho = sim.create_ode(length=2, needs_nbody=True)
        ode_ho.derivatives = derivatives_ho

        ode_ho.y[0] = 1. 
        ode_ho.y[1] = 0. # zero velocity

        sim.integrate(20.*math.pi)
        self.assertLess(math.fabs(ode_ho.y[0]-1.),2e-10)
        self.assertLess(math.fabs(ode_ho.y[1]),2e-9)
        e1 = sim.energy()
        self.assertLess(math.fabs((e0-e1)/e0),1e-10)

class TestIntegratorTrace(unittest.TestCase):

    def test_no_effect_tp(self):
        # tests if test particle encounters have an effect
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1)
        sim.move_to_com()
        sim.N_active=2
        sim.integrator = "trace"
        sim.ri_trace.peri_mode = 2
        sim.dt = 0.1
        sim2 = sim.copy()
        p = sim.particles[1].copy()
        p.x += 0.01
        p.m = 0
        sim.add(p)
        sim.step()
        sim2.step()

        self.assertEqual(sim.particles[1].x,sim2.particles[1].x)
        self.assertEqual(sim.particles[1].vx,sim2.particles[1].vx)
        self.assertEqual(sim.particles[0].x,sim2.particles[0].x)
        self.assertEqual(sim.particles[0].vx,sim2.particles[0].vx)

    def test_outer_solar(self):
        sim = rebound.Simulation()
        rebound.data.add_outer_solar_system(sim)

        sim.integrator = "trace"
        sim.ri_trace.peri_mode = 2
        P = sim.particles[1].P
        sim.dt = 1e-3*P

        E0 = sim.energy()
        sim.integrate(1000)
        dE = abs((sim.energy() - E0)/E0)
        self.assertLess(dE,2e-10)

    def test_order_doesnt_matter_tp0(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1.1,e=0.1,primary=sim.particles[0])
        sim.add(m=5e-3,a=1.0,e=0.1,primary=sim.particles[0])

        sim.integrator = "trace"
        sim.ri_trace.peri_mode = 2
        #sim.integrator = "whfast"
        #sim.ri_whfast.coordinates = "democraticheliocentric"
        sim.dt = 1e-2*2.*3.14
        sim.N_active = 1
        sim.testparticle_type = 0

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            sim.integrate(1000)
            self.assertEqual(1,len(w))

        o1 = sim.particles[1].orbit(primary=sim.particles[0])
        o2 = sim.particles[2].orbit(primary=sim.particles[0])


        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=5e-3,a=1.0,e=0.1,primary=sim.particles[0])
        sim.add(m=1e-3,a=1.1,e=0.1,primary=sim.particles[0])

        sim.integrator = "trace"
        sim.ri_trace.peri_mode = 2
        sim.dt = 1e-2*2.*3.14
        sim.N_active = 1
        sim.testparticle_type = 0

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            sim.integrate(1000)
            self.assertEqual(1,len(w))

        o3 = sim.particles[1].orbit(primary=sim.particles[0])
        o4 = sim.particles[2].orbit(primary=sim.particles[0])

        self.assertLess(abs(o1.e-o4.e),2e-16)
        self.assertLess(abs(o2.e-o3.e),2e-16)
        self.assertLess(abs(o1.a-o4.a),2e-16)
        self.assertLess(abs(o2.a-o3.a),2e-16)

    def test_order_doesnt_matter_tp1(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1.1,e=0.1,primary=sim.particles[0])
        sim.add(m=5e-3,a=1.0,e=0.1,primary=sim.particles[0])

        sim.integrator = "trace"
        sim.ri_trace.peri_mode = 2
        sim.dt = 1e-2*2.*3.14
        sim.N_active = 1
        sim.testparticle_type = 1

        sim.integrate(1000)

        o1 = sim.particles[1].orbit(primary=sim.particles[0])
        o2 = sim.particles[2].orbit(primary=sim.particles[0])


        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=5e-3,a=1.0,e=0.1,primary=sim.particles[0])
        sim.add(m=1e-3,a=1.1,e=0.1,primary=sim.particles[0])

        sim.integrator = "trace"
        sim.ri_trace.peri_mode = 2
        sim.dt = 1e-2*2.*3.14
        sim.N_active = 1
        sim.testparticle_type = 1

        sim.integrate(1000)

        o3 = sim.particles[1].orbit(primary=sim.particles[0])
        o4 = sim.particles[2].orbit(primary=sim.particles[0])

        self.assertLess(abs(o1.e-o4.e),2e-13)
        self.assertLess(abs(o2.e-o3.e),9e-14)
        self.assertLess(abs(o1.a-o4.a),9e-14)
        self.assertLess(abs(o2.a-o3.a),4e-14)

    def test_outer_solar_massive(self):
        sim = rebound.Simulation()
        rebound.data.add_outer_solar_system(sim)
        for i in range(1,sim.N):
            sim.particles[i].m *=50.

        sim.integrator = "trace"
        sim.ri_trace.peri_mode = 2
        P = sim.particles[1].P
        sim.dt = 1e-3*P

        E0 = sim.energy()
        sim.integrate(1000)
        dE = abs((sim.energy() - E0)/E0)
        self.assertLess(dE,7e-8)

    def test_simple_collision(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1e-5,r=1.6e-4,a=0.5,e=0.1)    #these params lead to collision on my machine
        sim.add(m=1e-8,r=4e-5,a=0.55,e=0.4,f=-0.94)
        mtot0= sum([p.m for p in sim.particles])
        com0 = sim.com()
        N0 = sim.N

        sim.integrator = "trace"
        sim.ri_trace.peri_mode = 2
        sim.dt = 0.01
        sim.track_energy_offset = 1
        sim.collision = "direct"
        sim.collision_resolve = "merge"

        E0 = sim.energy()
        sim.integrate(1)
        com1 = sim.com()
        self.assertAlmostEqual(com1.vx,com0.vx,delta=1e-16)
        self.assertAlmostEqual(com1.vy,com0.vy,delta=1e-16)
        self.assertAlmostEqual(com1.vz,com0.vz,delta=1e-16)
        mtot1= sum([p.m for p in sim.particles])
        self.assertEqual(mtot0,mtot1)
        dE = abs((sim.energy() - E0)/E0)
        self.assertLess(dE,3e-9)
        self.assertEqual(N0-1,sim.N)


    def test_planetesimal_collision(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1e-5,r=1.6e-4,a=0.5,e=0.1)    #these params lead to collision on my machine
        sim.N_active = 2
        sim.add(m=1e-8,r=4e-5,a=0.55,e=0.4,f=-0.94)
        mtot0= sum([p.m for p in sim.particles])
        N0 = sim.N

        sim.integrator = "trace"
        sim.ri_trace.peri_mode = 2
        sim.dt = 0.01
        sim.testparticle_type = 1
        sim.track_energy_offset = 1
        sim.collision = "direct"
        sim.collision_resolve = "merge"

        E0 = sim.energy()
        sim.integrate(1)
        mtot1= sum([p.m for p in sim.particles])
        self.assertEqual(mtot0,mtot1)
        dE = abs((sim.energy() - E0)/E0)
        self.assertLess(dE,3e-9)
        self.assertEqual(N0-1,sim.N)

    def test_massive_ejection(self):
        sim = rebound.Simulation()
        sim.add(m=1.)
        sim.add(m=1e-4,r=1.6e-4,a=0.5,e=0.1)
        sim.add(m=1e-6,r=4e-5,a=0.6)
        sim.particles[2].vy *= 2

        sim.N_active = 2

        sim.integrator = "trace"
        sim.ri_trace.peri_mode = 2
        sim.dt = 0.01
        sim.testparticle_type = 1
        sim.collision = "direct"
        sim.collision_resolve = "merge"
        sim.track_energy_offset = 1

        sim.boundary = "open"
        boxsize = 3.
        sim.configure_box(boxsize)

        E0 = sim.energy()
        sim.integrate(1)
        dE = abs((sim.energy() - E0)/E0)
        self.assertLess(dE,4e-6)

    def test_collision_with_star_simple(self):
        sim = rebound.Simulation()
        sim.add(m=1.,r=1.)
        sim.add(m=1e-3,r=1.e-3,a=0.5)
        mtot0 = sum([p.m for p in sim.particles])
        com0 = sim.com()
        N0 = sim.N

        sim.integrator = "trace"
        sim.ri_trace.peri_mode = 2
        sim.dt = 0.01
        sim.track_energy_offset = 1
        sim.collision = "direct"
        sim.collision_resolve = "merge"

        E0 = sim.energy()
        sim.integrate(1)
        com1 = sim.com()
        self.assertAlmostEqual(com1.vx,com0.vx,delta=1e-16)
        self.assertAlmostEqual(com1.vy,com0.vy,delta=1e-16)
        self.assertAlmostEqual(com1.vz,com0.vz,delta=1e-16)
        mtot1 = sum([p.m for p in sim.particles])
        self.assertEqual(mtot0,mtot1)
        self.assertEqual(N0-1,sim.N)
        dE = abs((sim.energy() - E0)/E0)
        self.assertLess(dE,1e-16)

    def test_collision_with_star(self):
        sim = rebound.Simulation()
        sim.add(m=1.,r=0.00465)
        sim.add(m=1e-5,r=1.6e-4,a=0.5,e=0.1,f=2.3)
        sim.add(m=1e-4,r=1.4e-3,x=1.,vx=-0.4) # falling onto the star
        sim.add(m=1e-5,r=1.6e-4,a=1.5,e=0.1)
        mtot0 = sum([p.m for p in sim.particles])
        com0 = sim.com()
        N0 = sim.N

        sim.integrator = "trace"
        sim.ri_trace.peri_mode = 2
        sim.dt = 0.01
        sim.track_energy_offset = 1
        sim.collision = "direct"
        sim.collision_resolve = "merge"
        sim.ri_trace.peri_crit_distance=0.2

        E0 = sim.energy()
        sim.integrate(1)
        com1 = sim.com()
        self.assertAlmostEqual(com1.vx,com0.vx,delta=1e-16)
        self.assertAlmostEqual(com1.vy,com0.vy,delta=1e-16)
        self.assertAlmostEqual(com1.vz,com0.vz,delta=1e-16)
        mtot1 = sum([p.m for p in sim.particles])
        self.assertEqual(mtot0,mtot1)
        self.assertEqual(N0-1,sim.N)
        dE = abs((sim.energy() - E0)/E0)
        # bad energy conservation due to democratic heliocentric!
        # worse than MERCURIUS, but still acceptable. Can maybe be improved...
        self.assertLess(dE,5e-2)

    def test_many_encounters(self):
        def get_sim():
            sim = rebound.Simulation()
            sim.add(m=1)
            # Setup using xyz instead of orbital elements for
            # machine independent test
            sim.add(m=0.0001,x=0.90000, y=0.00000, vx=0.00000, vy=1.10360)
            sim.add(m=0.0001, x=-1.17676, y=-0.05212, vx=0.22535, vy=-0.90102)
            sim.add(m=0.0001, x=-1.66025, y=-0.69852, vx=0.18932, vy=-0.60030)
            sim.add(m=0.0001, x=0.57904, y=1.03836, vx=-0.69267, vy=0.75995)
            sim.add(m=0.0001, x=-0.41683, y=0.83128, vx=-1.03478, vy=-0.72482)
            sim.add(m=0.0001, x=1.83969, y=0.32938, vx=-0.55114, vy=0.51646)
            sim.move_to_com()
            sim.dt = 0.034
            return sim

        sim = get_sim()
        sim.integrator = "trace"
        sim.ri_trace.peri_mode = 2
        E0 = sim.energy()
        start=datetime.now()
        sim.integrate(2000)
        time_trace = (datetime.now()-start).total_seconds()
        dE_trace = abs((sim.energy() - E0)/E0)

        sim = get_sim()
        sim.integrator = "ias15"
        start=datetime.now()
        sim.integrate(2000)
        time_ias15 = (datetime.now()-start).total_seconds()
        dE_ias15 = abs((sim.energy() - E0)/E0)

        sim = get_sim()
        sim.integrator = "whfast"
        start=datetime.now()
        sim.integrate(2000)
        time_whfast = (datetime.now()-start).total_seconds()
        dE_whfast = abs((sim.energy() - E0)/E0)

        # Note: precision might vary on machine as initializations use cos/sin
        # and are therefore machine dependent.
        self.assertLess(dE_trace,5e-5)              # reasonable precision for trace. Changed by Hanno 23 Jan 2024
        self.assertLess(dE_trace/dE_whfast,1e-4)    # at least 1e4 times better than whfast
        self.assertLess(time_trace,time_ias15) # faster than ias15
        if sys.maxsize > 2**32: # 64 bit
            self.assertEqual(7060.644251181158, sim.particles[5].x) # Check if bitwise unchanged

    # TLu additional tests
    def test_chaotic_exchange(self):
        def jacobi(sim):
            ps = sim.particles
            star = ps[0]
            planet = ps[1]
            particle = ps[2]
            rstar = np.array(star.xyz)
            rplanet = np.array(planet.xyz)
            r = np.array(particle.xyz)
            v = np.array(particle.vxyz)

            KE = 0.5 * v@v # test particle kinetic energy
            mu1 = sim.G * star.m
            mu2 = sim.G * planet.m
            r1 = r-rstar
            r2 = r-rplanet
            PE = -1*mu1/np.sqrt(r1@r1) - mu2/np.sqrt(r2@r2) # test particle potential energy

            lz = np.cross(r,v)[-1]

            CJ = 2 * planet.n * lz - 2 * (KE + PE) # jacobi constant
            return CJ

        sim = chaotic_exchange_sim()
        sim.integrator = "trace"
        sim.ri_trace.peri_mode = 2
        sim.ri_trace.r_crit_hill *= 1.21 # previously this was hardcoded
        sim.dt = (8./365.)*2.*math.pi
        E0 = jacobi(sim)
        start=datetime.now()
        sim.integrate(5000.)
        time_trace = (datetime.now()-start).total_seconds()
        dE_trace = abs((jacobi(sim) - E0)/E0)

        self.assertLess(dE_trace, 3e-6)              # reasonable precision for trace

if __name__ == "__main__":
    unittest.main()



import rebound
import unittest
from ctypes import c_size_t, sizeof

class TestSizeOfSimmulation(unittest.TestCase):
    
    def test_size_of_simulation(self):
        cl = rebound.clibrebound
        cl.reb_simulation_struct_size.res_type = c_size_t
        simulation_size_c = cl.reb_simulation_struct_size()

        self.assertEqual(simulation_size_c, sizeof(rebound.Simulation))

if __name__ == "__main__":
    unittest.main()



import ctypes


IAS15_ADAPTIVE_MODES = {"individual": 0, "global": 1, "prs23": 2, "aarseth85": 3}
class reb_dp7(ctypes.Structure):
    _fields_ = [("p0", ctypes.POINTER(ctypes.c_double)),
                ("p1", ctypes.POINTER(ctypes.c_double)),
                ("p2", ctypes.POINTER(ctypes.c_double)),
                ("p3", ctypes.POINTER(ctypes.c_double)),
                ("p4", ctypes.POINTER(ctypes.c_double)),
                ("p5", ctypes.POINTER(ctypes.c_double)),
                ("p6", ctypes.POINTER(ctypes.c_double))]

class IntegratorIAS15(ctypes.Structure):
    """
    This class is an abstraction of the C-struct reb_integrator_ias15.
    It controls the behaviour of the SEI integrator. See Rein & Spiegel (2015)
    for more information.

    Example usage:

    >>> sim = rebound.Simulation()
    >>> sim.integrator = "ias15"
    >>> sim.ri_ias15.epsilon = 0.

    :ivar float epsilon:          
        Controls the precision of the integrator. Set to 0 for fixed timesteps.
    
    :ivar float min_dt:          
        IAS15 is an adaptive method. This sets the minimum timestep.
    
    :ivar float adaptive_mode:
        Determines how the adaptive timestep is chosen. 
        This replaces the previous epsilon_global variable.
        The default is 2 which corresponds to the timestep criterion described in Pham, Rein, and Spiegel (2024).  This should be optimal in almost all cases.
        If set to 0, the fractional error is estimated via `max(acceleration_error/acceleration)` and the timestep criterion of Rein and Spiegel (2015) is used. 
        If set to 1, IAS15 estimates the fractional error via `max(acceleration_error)/max(acceleration)` where the maximum is taken over all particles. As before, the timestep criterion of Rein and Spiegel (2015) is used. This was the default until January 2024. 
        If set to 3, then the criterion of Aarseth (1985) is used.
    
    """
    def __repr__(self):
        return '<{0}.{1} object at {2}, epsilon={3}, min_dt={4}>'.format(self.__module__, type(self).__name__, hex(id(self)), self.epsilon, self.min_dt)
    
    @property
    def adaptive_mode(self):
        """
        Get or set the adaptive mode that determines how the timestep is chosen in IAS15.

        Available adaptive_modes are:

        - ``'individual'`` (timescale estimate based on individual particles)
        - ``'global'`` (used to be the default until 01/2024, global estimate)
        - ``'PRS23'`` (The default. Described in Pham, Rein, Spiegel 2023)
        - ``'Aarseth85'`` (Aarseth 1985 algorithm)
        """
        i = self._adaptive_mode
        for name, _i in IAS15_ADAPTIVE_MODES.items():
            if i==_i:
                return name
        return i
    @adaptive_mode.setter
    def adaptive_mode(self, value):
        if isinstance(value, int):
            self._adaptive_mode = ctypes.c_uint(value)
        elif isinstance(value, basestring):
            value = value.lower().replace(" ", "")
            if value in IAS15_ADAPTIVE_MODES: 
                self._adaptive_mode = IAS15_ADAPTIVE_MODES[value]
            else:
                raise ValueError("Warning. Kernel not found.")
    
    _fields_ = [("epsilon", ctypes.c_double),
                ("min_dt", ctypes.c_double),
                ("_adaptive_mode", ctypes.c_uint),
                ("_iterations_max_exceeded", ctypes.c_uint64),
                ("_N_allocated", ctypes.c_uint),
                ("_at", ctypes.POINTER(ctypes.c_double)),
                ("_x0", ctypes.POINTER(ctypes.c_double)),
                ("_v0", ctypes.POINTER(ctypes.c_double)),
                ("_a0", ctypes.POINTER(ctypes.c_double)),
                ("_csx", ctypes.POINTER(ctypes.c_double)),
                ("_csv", ctypes.POINTER(ctypes.c_double)),
                ("_csa0", ctypes.POINTER(ctypes.c_double)),
                ("_g", reb_dp7),
                ("_b", reb_dp7),
                ("_csb", reb_dp7),
                ("_e", reb_dp7),
                ("_br", reb_dp7),
                ("_er", reb_dp7),
                ("_map", ctypes.POINTER(ctypes.c_int)),
                ("_map_allocated_n", ctypes.c_uint),
                ]



import ctypes

SABA_TYPES = {
        "1": 0x0, "2": 0x1, "3": 0x2, "4": 0x3,
        "cm1": 0x100, "cm2": 0x101, "cm3": 0x102, "cm4": 0x103,
        "cl1": 0x200, "cl2": 0x201, "cl3": 0x202, "cl4": 0x203,
        "10,4": 0x4, "8,6,4": 0x5, "10,6,4": 0x6,
        "h8,4,4": 0x7, "h8,6,4": 0x8, "h10,6,4": 0x9,
        }

class IntegratorSABA(ctypes.Structure):
    """
    This class is an abstraction of the C-struct reb_integrator_saba.
    It controls the behaviour of the SABA integrator family.
    See Rein, Tamayo, and Brown (2019) for more details.

    :ivar str type:      
        Set the type of SABA integrator manually. The type can also be set by setting
        the integrator field in the REBOUND simulation. 

    :ivar int safe_mode:      
        This variable acts the same as for WHFast.
        If safe_mode is 1 (default) particles can be modified between
        timesteps and particle velocities and positions are always synchronised.
        If you set safe_mode to 0, the speed and accuracy of the integrator will improve.
        However, make sure you are aware of the consequences. Read the iPython tutorial
        on advanced WHFast usage to learn more.
   
    Example usage:
    
    >>> sim = rebound.Simulation()
    >>> sim.integrator = "SABA(10,6,4)"
    
    >>> sim = rebound.Simulation()
    >>> sim.integrator = "SABA"
    >>> sim.ri_saba.type = "(10,6,4)"
    
    >>> sim = rebound.Simulation()
    >>> sim.integrator = "SABACL4"
    >>> sim.ri_saba.safe_mode = 0

    """
    _fields_ = [("_type", ctypes.c_uint),
                ("safe_mode", ctypes.c_uint),
                ("is_synchronized", ctypes.c_uint),
                ("keep_unsynchronized", ctypes.c_uint),
            ]
    @property
    def type(self):
        """
        Get or set the type of SABA integrator.
        """
        i = self._type
        for name, _i in SABA_TYPES.items():
            if i==_i:
                return name
        return i
    @type.setter
    def type(self, value):
        if isinstance(value, int):
            self._type = value
        elif isinstance(value, basestring):
            value = value.lower().replace(" ", "").replace("(", "").replace(")", "")
            if value in SABA_TYPES: 
                self._type = SABA_TYPES[value]
            else:
                raise ValueError("Warning. SABA type not found.")




import ctypes

from ..particle import Particle

class IntegratorWHFast512(ctypes.Structure):
    _fields_ = [("gr_potential", ctypes.c_uint),
                ("N_systems", ctypes.c_uint),
                ("keep_unsynchronized", ctypes.c_uint),
                ("is_synchronized", ctypes.c_uint),
                ("_N_allocated", ctypes.c_uint),
                ("recalculate_constants", ctypes.c_uint),
                ("_p_jh", ctypes.POINTER(Particle)),
                ("_p_jh0", Particle*4)]



# -*- coding: utf-8 -*-
"""An N-body integrator package for python."""

import sys
import os
import warnings
import platform
from ctypes import cdll, c_char_p

# Find suffix
if platform.system()=="Windows" and sys.version_info.major<=3 and sys.version_info.minor<8:
    # Using distutils.sysconfig instead of sysconfig because 
    # of a bug in Python < 3.8 on windows
    import distutils.sysconfig as sysconfig
else:
    import sysconfig
suffix = sysconfig.get_config_var('EXT_SUFFIX')

if suffix is None:
    suffix = ".so"

try: # Only needed for pyodide
    import pyodide_js
    from site import getsitepackages
    pyodide_js._module.loadDynamicLibrary(getsitepackages()[0]+"/librebound"+suffix)
    del getsitepackages
    del pyodide_js
except:
    pass


# Make changes for python 2 and 3 compatibility
try:
    import builtins      # if this succeeds it's python 3.x
    builtins.xrange = range
    builtins.basestring = (str,bytes)
except ImportError:
    pass                 # python 2.x


# Import shared library
pymodulepath = os.path.dirname(os.path.abspath(__file__))
pymodulepath = os.path.abspath(os.path.join(pymodulepath, os.pardir))
__libpath__ = os.path.join(pymodulepath, "librebound"+suffix)
clibrebound = cdll.LoadLibrary(__libpath__)

# Version
__version__ = c_char_p.in_dll(clibrebound, "reb_version_str").value.decode('ascii')

# Build
__build__ = c_char_p.in_dll(clibrebound, "reb_build_str").value.decode('ascii')

# Githash
__githash__ = c_char_p.in_dll(clibrebound, "reb_githash_str").value.decode('ascii')

# Check for version
moduleversion = sys.modules["rebound"].__version__
libreboundversion = __version__
if moduleversion != libreboundversion:
    warnings.warn("WARNING: python module and librebound have different version numbers: '%s' vs '%s'.\n" %(moduleversion, libreboundversion), ImportWarning)
        
# Exceptions
class GenericError(Exception):
    """The simulation exited with a generic error."""
    pass

class Encounter(Exception):
    """The simulation exited because a close encounter has been detected.
    You may want to search for the pair of bodies which have the smallest distance."""
    pass

class Collision(Exception):
    """The simulation exited because a collision has been detected.
    You may want to search for which particles have a last_collision time equal to the simulation time."""
    pass

class Escape(Exception):
    """The simulation exited because a particle has been se encounter has been detected.
    You may want to search for the particle with the largest distance from the
    origin and remove it from the simulation."""
    pass

class NoParticles(Exception):
    """The simulation exited because no particles are left in the simulation."""
    pass

class ParticleNotFound(Exception):
    """Particle was not found in the simulation."""
    pass

from .hash import hash
from .tools import mod2pi, M_to_f, E_to_f, M_to_E, spherical_to_xyz, xyz_to_spherical
from .simulation import Simulation, Variation, ODE, Vec3d, Vec3dBasic, CollisionS # CollisionS is the collision struct, not the exception
from .rotation import Rotation
from .orbit import Orbit
from .particle import Particle
from .plotting import OrbitPlot, OrbitPlotSet
from .simulationarchive import Simulationarchive

__all__ = ["__libpath__", "__version__", "__build__", "__githash__", "Simulationarchive", "Simulation", "Orbit", "OrbitPlot", "OrbitPlotSet", "Particle", "GenericError", "Encounter", "Collision", "CollisionS", "Escape", "NoParticles", "ParticleNotFound", "Variation", "clibrebound", "mod2pi", "M_to_f", "E_to_f", "M_to_E", "ODE", "Rotation", "Vec3d", "spherical_to_xyz", "xyz_to_spherical"]



import rebound
import unittest
import os
import warnings
import urllib.request


class TestServer(unittest.TestCase):
    def test_start_without_download(self):
        with open("rebound.html", "w") as f:
            f.write("<h1>Hi</h1>")
        sim = rebound.Simulation()
        sim.start_server(port=1234)
        self.assertNotEqual(sim._server_data,None)
        self.assertEqual(sim._server_data.contents.ready,1)
        self.assertEqual(sim._server_data.contents.port,1234)
    
    def test_start_with_download(self):
        if os.path.isfile("rebound.html"):
            os.remove("rebound.html")
        sim = rebound.Simulation()
        with warnings.catch_warnings(record=True) as w: 
            warnings.simplefilter("always")
            sim.start_server(port=1234)
            self.assertGreater(len(w),0)
        self.assertNotEqual(sim._server_data,None)
        self.assertEqual(sim._server_data.contents.ready,1)
        self.assertEqual(sim._server_data.contents.port,1234)
    
    def test_connect(self):
        teststring = "<h1>Hi</h1>"
        with open("rebound.html", "w") as f:
            f.write(teststring )
        sim = rebound.Simulation()
        sim.start_server(port=1234)
        contents = urllib.request.urlopen("http://localhost:1234/").read()
        contents = contents.decode("ascii")
        self.assertEqual(contents, teststring)

    def test_pause(self):
        teststring = "<h1>Hi</h1>"
        with open("rebound.html", "w") as f:
            f.write(teststring )
        sim = rebound.Simulation()
        sim.start_server(port=1234)
        sim._status = -1;
        contents = urllib.request.urlopen("http://localhost:1234/keyboard/32")
        self.assertEqual(sim._status, -3)

    
    
if __name__ == "__main__":
    unittest.main()



import ctypes
from . import GenericError
from .particle import Particle
from .simulation import Simulation

class Variation(ctypes.Structure):
    """
    REBOUND Variational Configuration Object.

    This object encapsulated the configuration of one set of variational 
    equations in a REBOUND simulation.  It is an abstraction of the 
    C struct reb_variational_configuration.

    None of the fields in this struct should be changed after it has
    been initialized.

    One rebound simulation can include any number of first and second order 
    variational equations.

    Note that variations are only encoded as particles for convenience.  
    A variational particle's position and velocity should be interpreted as a derivative, i.e. how much that position or velocity varies with respect to the first or second-order variation.  
    See ipython_examples/VariationalEquations.ipynb and Rein and Tamayo (2016) for details.

    Examples
    --------

    >>> sim = rebound.Simulation()          # Create a simulation
    >>> sim.add(m=1.)                       # Add a star
    >>> sim.add(m=1.e-3, a=1.)              #     a planet
    >>> var_config = sim.add_variation()    # Add a set of variational equations. 
    >>> var_config.particles[1].x = 1.      # Initialize the variational particle corresponding to the planet
    >>> sim.integrate(100.)                 # Integrate the simulation
    >>> print(var_config.particles[0].vx)   # Print the velocity of the variational particle corresponding to the star
    """
    _fields_ = [
                ("_sim", ctypes.POINTER(Simulation)),
                ("order", ctypes.c_int),
                ("index", ctypes.c_int),
                ("testparticle", ctypes.c_int),
                ("index_1st_order_a", ctypes.c_int),
                ("index_1st_order_b", ctypes.c_int),
                ("_lrescale", ctypes.c_double)]

    def vary(self, particle_index, variation, variation2=None, primary=None):
        """
        This function can be used to initialize the variational particles that are 
        part of a Variation.
    
        Note that rather than using this convenience function, one can 
        also directly manipulate the particles' coordinate using the following
        syntax:

        >>> var = sim.add_variation()
        >>> var.particles[0].x = 1.

        The ``vary()`` function is useful for initializing variations corresponding to 
        changes in one of the orbital parameters for a particle on a bound 
        Keplerian orbit.

        The function supports both first and second order variations in the following
        classical orbital parameters:
          a, e, inc, omega, Omega, f
        as well as the Pal (2009) coordinates: 
          a, h, k, ix, iy, lambda
        and in both cases the mass m of the particle. The advantage of the Pal coordinate
        system is that all derivatives are well behaved (infinitely differentiable).
        Classical orbital parameters on the other hand exhibit coordinate singularities, 
        for example when e=0.
        
        The following example initializes the variational particles corresponding to a 
        change in the semi-major axis of the particle with index 1:
        
        >>> var = sim.add_variation()
        >>> var.vary(1,"a")

        Parameters
        ----------
        particle_index : int
            The index of the particle that should be varied. The index starts at 0 and runs through N-1. The first particle added to a simulation receives the index 0, the second 1, and the on.
        variation : string
            This parameter determines which orbital parameter is varied. 
        variation2: string, optional
            This is only used for second order variations which can depend on two varying parameters. If omitted, then it is assumed that the parameter variation is variation2.
        primary: Particle, optional
            By default, variational particles are created in the Heliocentric frame. 
            Set this parameter to use any other particles as a primary (e.g. the center of mass).
        """
        if self.order==2 and variation2 is None:
            variation2 = variation
        if self._sim is not None:
            sim = self._sim.contents
            particles = sim.particles
        else:
            raise RuntimeError("Something went wrong. Cannot seem to find simulation corresponding to variation.")
        if self.testparticle >= 0:
            particles[self.index] = Particle(simulation=sim,particle=particles[particle_index], variation=variation, variation2=variation2, primary=primary)
        else:
            particles[self.index + particle_index] = Particle(simulation=sim,particle=particles[particle_index], variation=variation, variation2=variation2, primary=primary)

    @property
    def particles(self):
        """
        Access the variational particles corresponding to this set of variational equations.

        The function returns a list of particles which are sorted in the same way as those in 
        sim.particles

        The particles are pointers and thus can be modified. 

        If there are N real particles, this function will also return a list of N particles (all of which are 
        variational particles). 
        """
        sim = self._sim.contents
        ps = []
        if self.testparticle>=0:
            N = 1
        else:
            N = sim.N-sim.N_var 
        
        ParticleList = Particle*N
        ps = ParticleList.from_address(ctypes.addressof(sim._particles.contents)+self.index*ctypes.sizeof(Particle))
        return ps
    
    @property
    def lrescale(self):
        """
        Access the lrescale parameter. 
        
        This is a property because sim.add_variation() returns a copy of the struct, so need to find up-to-date reb_variational_configuration struct in simulation.
        """
        sim = self._sim.contents
        for i in range(sim.N_var_config):
            if sim.var_config[i].index == self.index:
                return sim.var_config[i]._lrescale
        raise GenericError("An error occured while trying to find variational struct in simulation.")
    
    @lrescale.setter
    def lrescale(self, value):
        """
        Set the lrescale parameter. 
        
        This is a property because sim.add_variation() returns a copy of the struct, so need to find up-to-date reb_variational_configuration struct in simulation.
        """
        sim = self._sim.contents
        for i in range(sim.N_var_config):
            if sim.var_config[i].index == self.index:
                self._lrescale = ctypes.c_double(value)
                sim.var_config[i]._lrescale = ctypes.c_double(value)
                return

        raise GenericError("An error occured while trying to find variational struct in simulation.")




import rebound
import unittest


class TestIntegratorTraceBasic(unittest.TestCase):
   
    def test_trace_encounter_condition(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1)
        sim.add(m=1e-3,a=1.1,f=0.25) # First step will be rejected.
        sim.add(m=1e-3,a=3.1)
        sim.integrator = "TRACE"
        sim.ri_trace.r_crit_hill *= 1.21
        sim.dt = 0.14 

        for k in range(3):
            sim.step()
            self.assertEqual(sim.ri_trace._current_C,0)
            self.assertEqual(sim.ri_trace._encounter_N,3)
            self.assertEqual(sim.ri_trace._encounter_N_active,3)
            for i in range(sim.N):
                for j in range(i+1,sim.N):
                    if i==1 and j==2: 
                        self.assertEqual(sim.ri_trace._current_Ks[i*sim.N+j],1)
                    else:
                        self.assertEqual(sim.ri_trace._current_Ks[i*sim.N+j],0)
        
        # Change condition. No more encounters.
        sim.ri_trace.r_crit_hill = 0.1
        for k in range(3):
            sim.step()
            self.assertEqual(sim.ri_trace._current_C,0)
            self.assertEqual(sim.ri_trace._encounter_N,1)
            for i in range(sim.N):
                for j in range(i+1,sim.N):
                    self.assertEqual(sim.ri_trace._current_Ks[i*sim.N+j],0)

    def test_trace_encounter_prediction(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=9.55e-4,x=5.2)
        sim.add(x=5.3,y=0.36,vy=-7.2) # Non-encounter prediction misses this
        sim.integrator = "TRACE"
        sim.dt = 0.01
        sim.ri_trace.r_crit_hill = 1
        sim.step()
        
        self.assertEqual(sim.ri_trace._encounter_N,3)



if __name__ == "__main__":
    unittest.main()



import ctypes
from . import clibrebound

class BinaryFieldDescriptor(ctypes.Structure):
    """ 
    Describes the binary field in simulationarchives 

    Used here for unit tests only. Checking if ids are unique.
    """

    def __repr__(self):
        return '<{0}.{1} object at {2}, type={3}, dtype={4}, name=\'{5}\'>'.format(self.__module__, type(self).__name__, hex(id(self)), self.type, self.dtype, self.name.decode("ascii"))
    _fields_ = [("type", ctypes.c_uint),
                ("dtype", ctypes.c_int),
                ("name", ctypes.c_char*1024),
                ("offset", ctypes.c_size_t),
                ("offset_N", ctypes.c_size_t),
                ("element_size", ctypes.c_size_t),
                ]
def binary_field_descriptor_list():
    fd_pointer_t = ctypes.POINTER(BinaryFieldDescriptor)
    fd_pointer = (BinaryFieldDescriptor*3).in_dll(clibrebound, "reb_binary_field_descriptor_list")
    fd_pointer = ctypes.cast(fd_pointer, fd_pointer_t) # not sure why I have to do it this way
    l = []
    i=0
    while True:
        l.append(fd_pointer[i])
        if fd_pointer[i].name == b'end':
            break
        i += 1
    return l





import ctypes


EOS_TYPES = {
        "lf": 0x00,
        "lf4": 0x01,
        "lf6": 0x02,
        "lf8": 0x03,
        "lf4_2": 0x04,
        "lf8_6_4": 0x05,
        "plf7_6_4": 0x06,
        "pmlf4": 0x07,
        "pmlf6": 0x08,
        }


class IntegratorEOS(ctypes.Structure):
    """
    This class is an abstraction of the C-struct reb_integrator_eos.
    It controls the behaviour of the Embedded Operator Splitting methods. See Rein (2019) 
    for more details.
    
    :ivar int,string phi0      
        Sets the Phi_0 operator splitting method
    :ivar int,string phi1     
        Sets the Phi_1 operator splitting method
    :ivar int n     
        Sets the number of substeps taken by Phi_1
    :ivar int safe_mode  
        By default, safe_mode is on (1). Set to 0 (off) to combine
        drift step at the beginning and end of the Phi0 integrator steps.

    Example usage:
    
    >>> sim = rebound.Simulation()
    >>> sim.integrator = "eos"
    >>> sim.ri_eos.phi0 = "LF8_6_4"
    >>> sim.ri_eos.phi1 = "LF8"
    >>> sim.ri_eos.n = 1
    >>> sim.ri_eos.safe_mode = 0

    """
    def __repr__(self):
        return '<{0}.{1} object at {2}, safe_mode={3}, is_synchronized={4}, n={5}, phi0={6}, phi2={7}>'.format(self.__module__, type(self).__name__, hex(id(self)), self.safe_mode, self.is_synchronized, self.n, self.phi0, self.phi1)

    @property
    def phi0(self):
        """
        Get or set the type of operator splitting type for phi0.
        """
        i = self._phi0
        for name, _i in EOS_TYPES.items():
            if i==_i:
                return name
        return i
    @phi0.setter
    def phi0(self, value):
        if isinstance(value, int):
            self._phi0 = value
        elif isinstance(value, basestring):
            value = value.lower().replace(" ", "").replace("(", "").replace(")", "")
            if value in EOS_TYPES: 
                self._phi0 = EOS_TYPES[value]
            else:
                raise ValueError("Warning. EOS type %s not found."%value)
    @property
    def phi1(self):
        """
        Get or set the type of operator splitting type for phi1.
        """
        i = self._phi1
        for name, _i in EOS_TYPES.items():
            if i==_i:
                return name
        return i
    @phi1.setter
    def phi1(self, value):
        if isinstance(value, int):
            self._phi1 = value
        elif isinstance(value, basestring):
            value = value.lower().replace(" ", "").replace("(", "").replace(")", "")
            if value in EOS_TYPES: 
                self._phi1 = EOS_TYPES[value]
            else:
                raise ValueError("Warning. EOS type %s not found."%value)
    _fields_ = [
                ("_phi0",ctypes.c_uint),
                ("_phi1",ctypes.c_uint),
                ("n",ctypes.c_uint),
                ("safe_mode",ctypes.c_uint),
                ("is_synchronized",ctypes.c_uint),
                ]




import warnings
import unittest

class TestImport(unittest.TestCase):
    def test_no_warning_on_import(self):
        with warnings.catch_warnings(record=True) as w: 
            warnings.simplefilter("always")
            import rebound
            self.assertEqual(0,len(w))
    
if __name__ == "__main__":
    unittest.main()



import rebound
import unittest
from ctypes import c_int

class TestFPContract(unittest.TestCase):
    
    def test_fp_off(self):
        cl = rebound.clibrebound
        cl.reb_check_fp_contract.res_type = c_int
        fp_contract = cl.reb_check_fp_contract()

        # make sure floating point contraction are off
        self.assertEqual(fp_contract, 0)

if __name__ == "__main__":
    unittest.main()



import ctypes
from ..particle import Particle

WHFAST_KERNELS = {"default": 0, "modifiedkick": 1, "composition": 2, "lazy": 3}
WHFAST_COORDINATES = {"jacobi": 0, "democraticheliocentric": 1, "whds": 2, "barycentric": 3}

class IntegratorWHFast(ctypes.Structure):
    """
    This class is an abstraction of the C-struct reb_integrator_whfast.
    It controls the behaviour of the symplectic WHFast integrator described 
    in Rein and Tamayo (2015) and in Rein, Tamayo, and Brown (2019).
    
    This struct should be accessed via the simulation class only. Here is an 
    example:

    >>> sim = rebound.Simulation()
    >>> sim.integrator = "whfast"
    >>> sim.ri_whfast.corrector =  11
    >>> sim.ri_whfast.kernel = "lazy"

    
    :ivar int corrector:      
        The order of the symplectic corrector in the WHFast integrator.
        By default, the symplectic correctors are turned off (=0). For high
        accuracy simulation set this value to 11 or 17. For more details read 
        Rein and Tamayo (2015).
    :ivar int corrector2:      
        Second correctors (C2 of Wisdom et al 1996).
        By default, the second symplectic correctors are turned off (=0). 
        Set to 1 to turn them on.
    :ivar int/string kernel:      
        Kernel option. Set to 0 for the default WH kernel (standard kick step).
        Other options are "modifiedkick" (1), "composition" (2), "lazy" (3).
    :ivar int recalculate_coordinates_this_timestep:
        Sets a flag that tells WHFast that the particles have changed.
        Setting this flag to 1 (default 0) triggers the WHFast integrator
        to recalculate Jacobi/heliocenctric coordinates. This is needed 
        if the user changes the particle position, velocity or mass 
        in-between timesteps.  After every timestep the flag is set back 
        to 0, so if you continuously update the particles manually, 
        you need to set this flag to 1 after every timestep.
    :ivar string coordinates:
        Sets the internal coordinate system that WHFast is using. By default
        it uses ``'jacobi'`` (=0) coordinates. Other options are 
        ``'democraticheliocentric'`` (=1) and ``'whds'`` (=2). See Hernandez 
        and Dehnen (2017) for more information.
    :ivar int safe_mode:
        If safe_mode is 1 (default) particles can be modified between
        timesteps and particle velocities and positions are always synchronised.
        If you set safe_mode to 0, the speed and accuracy of WHFast improve.
        However, make sure you are aware of the consequences. Read the iPython tutorial
        on advanced WHFast usage to learn more.
    """
    _fields_ = [("corrector", ctypes.c_uint),
                ("corrector2", ctypes.c_uint),
                ("_kernel", ctypes.c_uint),
                ("_coordinates", ctypes.c_uint),
                ("recalculate_coordinates_this_timestep", ctypes.c_uint),
                ("safe_mode", ctypes.c_uint),
                ("keep_unsynchronized", ctypes.c_uint),
                ("_p_jh", ctypes.POINTER(Particle)),
                ("_p_temp", ctypes.POINTER(Particle)),
                ("is_synchronized", ctypes.c_uint),
                ("_N_allocated", ctypes.c_uint),
                ("_N_allocated_tmp", ctypes.c_uint),
                ("_timestep_warning", ctypes.c_uint),
                ("_recalculate_coordinates_but_not_synchronized_warning", ctypes.c_uint)]

    def __repr__(self):
        return '<{0}.{1} object at {2}, safe_mode={3}, keep_unsynchonized={4}, is_synchronized={5}, corrector={6}, corrector2={7}, kernel={8}>'.format(self.__module__, type(self).__name__, hex(id(self)), self.safe_mode, self.keep_unsynchronized, self.is_synchronized, self.corrector, self.corrector2, self.kernel)

    @property
    def coordinates(self):
        """
        Get or set the internal coordinate system.

        Available coordinate systems are:

        - ``'jacobi'`` (default)
        - ``'democraticheliocentric'``
        - ``'whds'``
        - ``'barycentric'``
        """
        i = self._coordinates
        for name, _i in WHFAST_COORDINATES.items():
            if i==_i:
                return name
        return i
    @coordinates.setter
    def coordinates(self, value):
        if isinstance(value, int):
            self._coordinates = ctypes.c_uint(value)
        elif isinstance(value, basestring):
            value = value.lower()
            if value in WHFAST_COORDINATES: 
                self._coordinates = WHFAST_COORDINATES[value]
            else:
                raise ValueError("Warning. Coordinate system not found.")
    @property
    def kernel(self):
        """
        Get or set the WHFast Kernel.

        Available kernels are:

        - ``'default'`` (standard WH kernel, kick)
        - ``'modifiedkick'`` (modified kick for newtonian gravity)
        - ``'composition'`` (Wisdom's composition method)
        - ``'lazy'`` (Lazy implementer's method)
        """
        i = self._kernel
        for name, _i in WHFAST_KERNELS.items():
            if i==_i:
                return name
        return i
    @kernel.setter
    def kernel(self, value):
        if isinstance(value, int):
            self._kernel = ctypes.c_uint(value)
        elif isinstance(value, basestring):
            value = value.lower().replace(" ", "")
            if value in WHFAST_KERNELS: 
                self._kernel = WHFAST_KERNELS[value]
            else:
                raise ValueError("Warning. Kernel not found.")





import rebound
import unittest

class TestBoundary(unittest.TestCase):
    
    def test_open(self):
        for gravity in ["basic", "tree"]:
            sim = rebound.Simulation()
            sim.boundary = "open"
            sim.gravity  = gravity
            sim.configure_box(10.)
            sim.add(m=0.1,x=1., vx=5.0)
            sim.add(m=0.1,x=-1., vx=-5.0)
            sim.add(m=0.1,y=1., vx=6.0)
            sim.add(m=0.1,x=-1., y=-1., vx=-3., vy=-3.)
            sim.add(m=0.1,z=1., vz=5.0)
            sim.add(m=0.1,z=-1., vz=-5.0)
            self.assertEqual(sim.N,6)
            sim.integrate(1.)
            self.assertEqual(sim.N,1)
            with self.assertRaises(rebound.NoParticles):
                sim.integrate(2.)
    
    def test_periodic(self):
        sim = rebound.Simulation()
        sim.boundary = "periodic"
        sim.configure_box(10.)
        sim.add(m=0.0,x=1., vx=5.0, vy=15.1, vz=26.)
        sim.add(m=0.0,x=-1., vx=-5.0, vy=-15.1, vz=-26.)
        sim.integrate(1.)
        self.assertAlmostEqual(sim.particles[0].x,-4,delta=1e-16)
        self.assertAlmostEqual(sim.particles[0].y,-4.9,delta=1e-14)
        self.assertAlmostEqual(sim.particles[0].z,-4.0,delta=1e-16)
        self.assertAlmostEqual(sim.particles[1].x,4,delta=1e-16)
        sim.integrate(2.)
        self.assertAlmostEqual(sim.particles[0].x,1,delta=1e-16)
        self.assertAlmostEqual(sim.particles[0].y,0.2,delta=1e-14)
        self.assertAlmostEqual(sim.particles[0].z,2.0,delta=1e-16)
        self.assertAlmostEqual(sim.particles[1].x,-1,delta=1e-16)
        self.assertEqual(sim.N,2)
    
    def test_shear_vertical(self):
        # Note: not physical
        sim = rebound.Simulation()
        sim.ri_sei.OMEGA = 1
        sim.configure_box(1)
        sim.N_ghost_x = 2
        sim.N_ghost_y = 2
        sim.integrator = "sei"
        sim.boundary   = "shear"
        sim.add(z=0.45,vz=100.0) 
        sim.add(z=-0.45,vz=-100.0) 
        sim.integrate(1)
        self.assertEqual(2,sim.N)

    def test_add_outside(self):
        for boundary in ["open", "shear", "periodic"]:
            sim = rebound.Simulation()
            sim.boundary = boundary
            sim.configure_box(10.)
            with self.assertRaises(RuntimeError):
                sim.add(x=6.)
            with self.assertRaises(RuntimeError):
                sim.add(x=-6.)
            with self.assertRaises(RuntimeError):
                sim.add(y=6.)
            with self.assertRaises(RuntimeError):
                sim.add(y=-6.)
            with self.assertRaises(RuntimeError):
                sim.add(z=6.)
            with self.assertRaises(RuntimeError):
                sim.add(z=-6.)
    
    
if __name__ == "__main__":
    unittest.main()



import rebound
import unittest
import sys
import warnings
import math
import random
from datetime import datetime

class TestTPOnlyEncounter(unittest.TestCase):
    
    def test_tponly_encounter0_mercurius(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-2,a=1)
        sim.add(m=1e-2,a=2)
        sim.move_to_com()
        sim.N_active = sim.N
        sim.integrator = "mercurius"
        sim.testparticle_type = 1 # Test particles can affect active particles
        sim.dt = 0.1
        sim2 = sim.copy()
        
        random.seed(10)
        for i in range(10):
            sim.add(a=1.0+1e-2*random.uniform(-1,1), f=0.3*random.uniform(-1,1))
            sim2.add(a=1.0+1e-2*random.uniform(-1,1), f=0.3*random.uniform(-1,1))


        sim.integrate(10)
        sim2.integrate(10)

        self.assertNotEqual(sim.particles[2].x, sim2.particles[2].x)
        self.assertNotEqual(sim.particles[1].x, sim2.particles[1].x)

    def test_tponly_encounter1_mercurius(self):
        sim = rebound.Simulation()
        sim.add(m=1)
        sim.add(m=1e-3,a=1)
        sim.add(m=1e-3,a=2)
        sim.move_to_com()
        sim.N_active = sim.N
        sim.integrator = "mercurius"
        sim.testparticle_type = 0 # Test particles can not affect active particles (default)
        sim.dt = 0.1
        sim2 = sim.copy()
        
        random.seed(10)
        for i in range(10):
            sim.add(a=1.0+1e-2*random.uniform(-1,1), f=0.3*random.uniform(-1,1))
            sim2.add(a=1.0+1e-2*random.uniform(-1,1), f=0.3*random.uniform(-1,1))


        sim.integrate(10)
        sim2.integrate(10)

        self.assertEqual(sim.particles[2].x, sim2.particles[2].x)
        self.assertEqual(sim.particles[1].x, sim2.particles[1].x)



if __name__ == "__main__":
    unittest.main()

